# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Unit tests for image operations
-------------------------------

Features from the "Operations" menu are covered by this test.
The "Operations" menu contains basic operations on images, such as
addition, multiplication, division, and more.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import Generator

import numpy as np
import pytest
import scipy.ndimage as spi
from guidata.qthelpers import qt_app_context

import cdl.computation.image as cpi
import cdl.obj
import cdl.param
from cdl.algorithms.datatypes import is_integer_dtype
from cdl.env import execenv
from cdl.tests.data import create_noisygauss_image
from cdl.utils.tests import check_array_result
from cdl.utils.vistools import view_images_side_by_side


def __iterate_images() -> Generator[cdl.obj.ImageObj, None, None]:
    """Iterate over all possible image types for testing."""
    size = 128
    for dtype in cdl.obj.ImageDatatypes:
        param = cdl.obj.NewImageParam.create(dtype=dtype, height=size, width=size)
        yield create_noisygauss_image(param, level=0.0)


def __iterate_image_couples() -> (
    Generator[tuple[cdl.obj.ImageObj, cdl.obj.ImageObj], None, None]
):
    """Iterate over all possible image couples for testing."""
    size = 128
    for dtype1 in cdl.obj.ImageDatatypes:
        param1 = cdl.obj.NewImageParam.create(dtype=dtype1, height=size, width=size)
        ima1 = create_noisygauss_image(param1, level=0.0)
        for dtype2 in cdl.obj.ImageDatatypes:
            param2 = cdl.obj.NewImageParam.create(dtype=dtype2, height=size, width=size)
            ima2 = create_noisygauss_image(param2, level=0.0)
            yield ima1, ima2


@pytest.mark.validation
def test_image_addition() -> None:
    """Image addition test."""
    execenv.print("*** Testing image addition:")
    for ima1, ima2 in __iterate_image_couples():
        dtype1, dtype2 = ima1.data.dtype, ima2.data.dtype
        execenv.print(f"  {dtype1} += {dtype2}: ", end="")
        exp = ima1.data.astype(float) + ima2.data.astype(float)
        cpi.compute_addition(ima2, ima1)
        check_array_result("Image addition", ima2.data, exp)


@pytest.mark.validation
def test_image_difference() -> None:
    """Image difference test."""
    execenv.print("*** Testing image difference:")
    for ima1, ima2 in __iterate_image_couples():
        dtype1, dtype2 = ima1.data.dtype, ima2.data.dtype
        execenv.print(f"  {dtype1} -= {dtype2}: ", end="")
        exp = ima1.data.astype(float) - ima2.data.astype(float)
        ima3 = cpi.compute_difference(ima1, ima2)
        check_array_result("Image difference", ima3.data, exp)


@pytest.mark.validation
def test_quadratic_difference() -> None:
    """Quadratic difference test."""
    execenv.print("*** Testing quadratic difference:")
    for ima1, ima2 in __iterate_image_couples():
        dtype1, dtype2 = ima1.data.dtype, ima2.data.dtype
        execenv.print(f"  ({dtype1} - {dtype2})/√2: ", end="")
        exp = (ima1.data.astype(float) - ima2.data.astype(float)) / np.sqrt(2)
        ima3 = cpi.compute_quadratic_difference(ima1, ima2)
        check_array_result("Image quadratic difference", ima3.data, exp)


@pytest.mark.validation
def test_image_product() -> None:
    """Image multiplication test."""
    execenv.print("*** Testing image multiplication:")
    for ima1, ima2 in __iterate_image_couples():
        dtype1, dtype2 = ima1.data.dtype, ima2.data.dtype
        execenv.print(f"  {dtype1} *= {dtype2}: ", end="")
        exp = ima1.data.astype(float) * ima2.data.astype(float)
        cpi.compute_product(ima2, ima1)
        check_array_result("Image multiplication", ima2.data, exp)


@pytest.mark.validation
def test_image_division() -> None:
    """Image division test."""
    execenv.print("*** Testing image division:")
    for ima1, ima2 in __iterate_image_couples():
        ima2.data = np.ones_like(ima2.data)
        dtype1, dtype2 = ima1.data.dtype, ima2.data.dtype
        execenv.print(f"  {dtype1} /= {dtype2}: ", end="")
        exp = ima1.data.astype(float) / ima2.data.astype(float)
        ima3 = cpi.compute_division(ima1, ima2)
        if not np.allclose(ima3.data, exp):
            with qt_app_context():
                view_images_side_by_side(
                    [ima1.data, ima2.data, ima3.data], ["ima1", "ima2", "ima3"]
                )
        check_array_result("Image division", ima3.data, exp)


def __constparam(value: float) -> cdl.param.ConstantParam:
    """Create a constant parameter."""
    return cdl.param.ConstantParam.create(value=value)


def __iterate_image_with_constant() -> (
    Generator[tuple[cdl.obj.ImageObj, cdl.param.ConstantParam], None, None]
):
    """Iterate over all possible image and constant couples for testing."""
    size = 128
    for dtype in cdl.obj.ImageDatatypes:
        param = cdl.obj.NewImageParam.create(dtype=dtype, height=size, width=size)
        ima = create_noisygauss_image(param, level=0.0)
        for value in (-1.0, 3.14, 5):
            p = __constparam(value)
            yield ima, p


@pytest.mark.validation
def test_image_addition_constant() -> None:
    """Image addition with constant test."""
    execenv.print("*** Testing image addition with constant:")
    for ima1, p in __iterate_image_with_constant():
        dtype1 = ima1.data.dtype
        execenv.print(f"  {dtype1} += constant ({p.value}): ", end="")
        expvalue = np.array(p.value).astype(dtype=dtype1)
        exp = ima1.data.astype(float) + expvalue
        ima2 = cpi.compute_addition_constant(ima1, p)
        check_array_result(f"Image + constant ({p.value})", ima2.data, exp)


@pytest.mark.validation
def test_image_difference_constant() -> None:
    """Image difference with constant test."""
    execenv.print("*** Testing image difference with constant:")
    for ima1, p in __iterate_image_with_constant():
        dtype1 = ima1.data.dtype
        execenv.print(f"  {dtype1} -= constant ({p.value}): ", end="")
        expvalue = np.array(p.value).astype(dtype=dtype1)
        exp = ima1.data.astype(float) - expvalue
        ima2 = cpi.compute_difference_constant(ima1, p)
        check_array_result(f"Image - constant ({p.value})", ima2.data, exp)


@pytest.mark.validation
def test_image_product_constant() -> None:
    """Image multiplication by constant test."""
    execenv.print("*** Testing image multiplication by constant:")
    for ima1, p in __iterate_image_with_constant():
        dtype1 = ima1.data.dtype
        execenv.print(f"  {dtype1} *= constant ({p.value}): ", end="")
        exp = ima1.data.astype(float) * p.value
        ima2 = cpi.compute_product_constant(ima1, p)
        check_array_result(f"Image x constant ({p.value})", ima2.data, exp)


@pytest.mark.validation
def test_image_division_constant() -> None:
    """Image division by constant test."""
    execenv.print("*** Testing image division by constant:")
    for ima1, p in __iterate_image_with_constant():
        dtype1 = ima1.data.dtype
        execenv.print(f"  {dtype1} /= constant ({p.value}): ", end="")
        exp = ima1.data.astype(float) / p.value
        ima2 = cpi.compute_division_constant(ima1, p)
        check_array_result(f"Image / constant ({p.value})", ima2.data, exp)


@pytest.mark.validation
def test_image_arithmetic() -> None:
    """Image arithmetic test."""
    execenv.print("*** Testing image arithmetic:")
    for ima1, ima2 in __iterate_image_couples():
        dtype1 = ima1.data.dtype
        p = cdl.param.ArithmeticParam.create()
        for o in p.operators:
            p.operator = o
            for a in (0.0, 1.0, 2.0):
                p.factor = a
                for b in (0.0, 1.0, 2.0):
                    p.constant = b
                    ima2.data = np.clip(ima2.data, 1, None)  # Avoid division by zero
                    ima3 = cpi.compute_arithmetic(ima1, ima2, p)
                    if o in ("×", "/") and a == 0.0:
                        exp = np.ones_like(ima1.data) * b
                    elif o == "+":
                        exp = np.add(ima1.data, ima2.data, dtype=float) * a + b
                    elif o == "×":
                        exp = np.multiply(ima1.data, ima2.data, dtype=float) * a + b
                    elif o == "-":
                        exp = np.subtract(ima1.data, ima2.data, dtype=float) * a + b
                    elif o == "/":
                        exp = np.divide(ima1.data, ima2.data, dtype=float) * a + b
                    if p.restore_dtype:
                        if is_integer_dtype(dtype1):
                            iinfo1 = np.iinfo(dtype1)
                            exp = np.clip(exp, iinfo1.min, iinfo1.max)
                        exp = exp.astype(dtype1)
                    check_array_result(
                        f"Arithmetic [{p.get_operation()}]", ima3.data, exp
                    )


@pytest.mark.validation
def test_image_abs() -> None:
    """Image absolute value test."""
    execenv.print("*** Testing image absolute value:")
    for ima1 in __iterate_images():
        execenv.print(f"  abs({ima1.data.dtype}): ", end="")
        exp = np.abs(ima1.data)
        ima2 = cpi.compute_abs(ima1)
        check_array_result("Image abs", ima2.data, exp)


@pytest.mark.validation
def test_image_re() -> None:
    """Image real part test."""
    execenv.print("*** Testing image real part:")
    for ima1 in __iterate_images():
        execenv.print(f"  re({ima1.data.dtype}): ", end="")
        exp = np.real(ima1.data)
        ima2 = cpi.compute_re(ima1)
        check_array_result("Image re", ima2.data, exp)


@pytest.mark.validation
def test_image_im() -> None:
    """Image imaginary part test."""
    execenv.print("*** Testing image imaginary part:")
    for ima1 in __iterate_images():
        execenv.print(f"  im({ima1.data.dtype}): ", end="")
        exp = np.imag(ima1.data)
        ima2 = cpi.compute_im(ima1)
        check_array_result("Image im", ima2.data, exp)


def __get_numpy_info(dtype: np.dtype) -> np.generic:
    """Get numpy info for a given data type."""
    if is_integer_dtype(dtype):
        return np.iinfo(dtype)
    return np.finfo(dtype)


@pytest.mark.validation
def test_image_astype() -> None:
    """Image type conversion test."""
    execenv.print("*** Testing image type conversion:")
    for ima1 in __iterate_images():
        for dtype_str in cpi.VALID_DTYPES_STRLIST:
            dtype1_str = str(ima1.data.dtype)
            execenv.print(f"  {dtype1_str} -> {dtype_str}: ", end="")
            dtype_exp = np.dtype(dtype_str)
            info_exp = __get_numpy_info(dtype_exp)
            info_ima1 = __get_numpy_info(ima1.data.dtype)
            if info_exp.min < info_ima1.min or info_exp.max > info_ima1.max:
                continue
            exp = np.clip(ima1.data, info_exp.min, info_exp.max).astype(dtype_exp)
            p = cdl.param.DataTypeIParam.create(dtype_str=dtype_str)
            ima2 = cpi.compute_astype(ima1, p)
            check_array_result(
                f"Image astype({dtype1_str}->{dtype_str})", ima2.data, exp
            )


@pytest.mark.validation
def test_image_exp() -> None:
    """Image exponential test."""
    execenv.print("*** Testing image exponential:")
    with np.errstate(over="ignore"):
        for ima1 in __iterate_images():
            execenv.print(f"  exp({ima1.data.dtype}): ", end="")
            exp = np.exp(ima1.data)
            ima2 = cpi.compute_exp(ima1)
            check_array_result("Image exp", ima2.data, exp)


@pytest.mark.validation
def test_image_log10() -> None:
    """Image base-10 logarithm test."""
    execenv.print("*** Testing image base-10 logarithm:")
    with np.errstate(over="ignore"):
        for ima1 in __iterate_images():
            execenv.print(f"  log10({ima1.data.dtype}): ", end="")
            exp = np.log10(np.exp(ima1.data))
            ima2 = cpi.compute_log10(cpi.compute_exp(ima1))
            check_array_result("Image log10", ima2.data, exp)


@pytest.mark.validation
def test_image_logp1() -> None:
    """Image log(1+n) test."""
    execenv.print("*** Testing image log(1+n):")
    with np.errstate(over="ignore"):
        for ima1 in __iterate_images():
            execenv.print(f"  log1p({ima1.data.dtype}): ", end="")
            p = cdl.param.LogP1Param.create(n=2)
            exp = np.log10(ima1.data + p.n)
            ima2 = cpi.compute_logp1(ima1, p)
            check_array_result("Image log1p", ima2.data, exp)


def __generic_flip_check(compfunc: callable, expfunc: callable) -> None:
    """Generic flip check function."""
    execenv.print(f"*** Testing image flip: {compfunc.__name__}")
    for ima1 in __iterate_images():
        execenv.print(f"  {compfunc.__name__}({ima1.data.dtype}): ", end="")
        ima2: cdl.obj.ImageObj = compfunc(ima1)
        check_array_result("Image flip", ima2.data, expfunc(ima1.data))


@pytest.mark.validation
def test_image_fliph() -> None:
    """Image horizontal flip test."""
    __generic_flip_check(cpi.compute_fliph, np.fliplr)


@pytest.mark.validation
def test_image_flipv() -> None:
    """Image vertical flip test."""
    __generic_flip_check(cpi.compute_flipv, np.flipud)


def __generic_rotate_check(angle: int) -> None:
    """Generic rotate check function."""
    execenv.print(f"*** Testing image {angle}° rotation:")
    for ima1 in __iterate_images():
        execenv.print(f"  rotate{angle}({ima1.data.dtype}): ", end="")
        ima2 = getattr(cpi, f"compute_rotate{angle}")(ima1)
        check_array_result(
            f"Image rotate{angle}", ima2.data, np.rot90(ima1.data, k=angle // 90)
        )


@pytest.mark.validation
def test_image_rotate90() -> None:
    """Image 90° rotation test."""
    __generic_rotate_check(90)


@pytest.mark.validation
def test_image_rotate270() -> None:
    """Image 270° rotation test."""
    __generic_rotate_check(270)


@pytest.mark.validation
def test_image_rotate() -> None:
    """Image rotation test."""
    execenv.print("*** Testing image rotation:")
    for ima1 in __iterate_images():
        for angle in (30, 45, 60, 120):
            execenv.print(f"  rotate{angle}({ima1.data.dtype}): ", end="")
            ima2 = cpi.compute_rotate(ima1, cdl.param.RotateParam.create(angle=angle))
            exp = spi.rotate(ima1.data, angle, reshape=False)
            check_array_result(f"Image rotate{angle}", ima2.data, exp)


if __name__ == "__main__":
    test_image_addition()
    test_image_product()
    test_image_division()
    test_image_difference()
    test_quadratic_difference()
    test_image_addition_constant()
    test_image_product_constant()
    test_image_difference_constant()
    test_image_division_constant()
    test_image_arithmetic()
    test_image_abs()
    test_image_re()
    test_image_im()
    test_image_astype()
    test_image_exp()
    test_image_log10()
    test_image_logp1()
    test_image_fliph()
    test_image_flipv()
    test_image_rotate90()
    test_image_rotate270()
    test_image_rotate()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Image processor object (see parent package :mod:`cdl.core.gui.processor`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from guidata.qthelpers import exec_dialog
from plotpy.widgets.resizedialog import ResizeDialog
from qtpy import QtWidgets as QW

import cdl.computation.base as cpb
import cdl.computation.image as cpi
import cdl.computation.image.detection as cpi_det
import cdl.computation.image.edges as cpi_edg
import cdl.computation.image.exposure as cpi_exp
import cdl.computation.image.morphology as cpi_mor
import cdl.computation.image.restoration as cpi_res
import cdl.computation.image.threshold as cpi_thr
import cdl.param
from cdl.algorithms.image import distance_matrix
from cdl.config import APP_NAME, Conf, _
from cdl.core.gui.processor.base import BaseProcessor
from cdl.core.gui.profiledialog import ProfileExtractionDialog
from cdl.core.model.base import ResultProperties, ResultShape
from cdl.core.model.image import ImageObj, ImageROI, ROI2DParam, create_image_roi
from cdl.utils.qthelpers import create_progress_bar, qt_try_except
from cdl.widgets import imagebackground

if TYPE_CHECKING:
    import guidata.dataset as gds


class ImageProcessor(BaseProcessor[ImageROI]):
    """Object handling image processing: operations, processing, analysis"""

    # pylint: disable=duplicate-code

    @qt_try_except()
    def compute_normalize(self, param: cpb.NormalizeParam | None = None) -> None:
        """Normalize data with :py:func:`cdl.computation.image.compute_normalize`"""
        self.compute_11(
            cpi.compute_normalize,
            param=param,
            paramclass=cpb.NormalizeParam,
            title=_("Normalize"),
        )

    @qt_try_except()
    def compute_sum(self) -> None:
        """Compute sum with :py:func:`cdl.computation.image.compute_addition`"""
        self.compute_n1("Σ", cpi.compute_addition, title=_("Sum"))

    @qt_try_except()
    def compute_addition_constant(self, param: cpb.ConstantParam | None = None) -> None:
        """Compute sum with a constant using
        :py:func:`cdl.computation.image.compute_addition_constant`"""
        self.compute_11(
            cpi.compute_addition_constant,
            param,
            paramclass=cpb.ConstantParam,
            title=_("Add constant"),
        )

    @qt_try_except()
    def compute_average(self) -> None:
        """Compute average with :py:func:`cdl.computation.image.compute_addition`
        and dividing by the number of images"""

        def func_objs(new_obj: ImageObj, old_objs: list[ImageObj]) -> None:
            """Finalize average computation"""
            new_obj.data = new_obj.data / float(len(old_objs))

        self.compute_n1(
            "μ", cpi.compute_addition, func_objs=func_objs, title=_("Average")
        )

    @qt_try_except()
    def compute_product(self) -> None:
        """Compute product with :py:func:`cdl.computation.image.compute_product`"""
        self.compute_n1("Π", cpi.compute_product, title=_("Product"))

    @qt_try_except()
    def compute_product_constant(self, param: cpb.ConstantParam | None = None) -> None:
        """Compute product with a constant using
        :py:func:`cdl.computation.image.compute_product_constant`"""
        self.compute_11(
            cpi.compute_product_constant,
            param,
            paramclass=cpb.ConstantParam,
            title=_("Product with constant"),
        )

    @qt_try_except()
    def compute_logp1(self, param: cdl.param.LogP1Param | None = None) -> None:
        """Compute base 10 logarithm using
        :py:func:`cdl.computation.image.compute_logp1`"""
        self.compute_11(cpi.compute_logp1, param, cpi.LogP1Param, title="Log10")

    @qt_try_except()
    def compute_rotate(self, param: cdl.param.RotateParam | None = None) -> None:
        """Rotate data arbitrarily using
        :py:func:`cdl.computation.image.compute_rotate`"""
        self.compute_11(cpi.compute_rotate, param, cpi.RotateParam, title="Rotate")

    @qt_try_except()
    def compute_rotate90(self) -> None:
        """Rotate data 90° with :py:func:`cdl.computation.image.compute_rotate90`"""
        self.compute_11(cpi.compute_rotate90, title="Rotate90")

    @qt_try_except()
    def compute_rotate270(self) -> None:
        """Rotate data 270° with :py:func:`cdl.computation.image.compute_rotate270`"""
        self.compute_11(cpi.compute_rotate270, title="Rotate270")

    @qt_try_except()
    def compute_fliph(self) -> None:
        """Flip data horizontally using
        :py:func:`cdl.computation.image.compute_fliph`"""
        self.compute_11(cpi.compute_fliph, title="HFlip")

    @qt_try_except()
    def compute_flipv(self) -> None:
        """Flip data vertically with :py:func:`cdl.computation.image.compute_flipv`"""
        self.compute_11(cpi.compute_flipv, title="VFlip")

    @qt_try_except()
    def distribute_on_grid(self, param: cdl.param.GridParam | None = None) -> None:
        """Distribute images on a grid"""
        title = _("Distribute on grid")
        edit, param = self.init_param(param, cpi.GridParam, title)
        if edit and not param.edit(parent=self.panel.parent()):
            return
        objs = self.panel.objview.get_sel_objects(include_groups=True)
        g_row, g_col, x0, y0, x0_0, y0_0 = 0, 0, 0.0, 0.0, 0.0, 0.0
        delta_x0, delta_y0 = 0.0, 0.0
        with create_progress_bar(self.panel, title, max_=len(objs)) as progress:
            for i_row, obj in enumerate(objs):
                progress.setValue(i_row + 1)
                QW.QApplication.processEvents()
                if progress.wasCanceled():
                    break
                if i_row == 0:
                    x0_0, y0_0 = x0, y0 = obj.x0, obj.y0
                else:
                    delta_x0, delta_y0 = x0 - obj.x0, y0 - obj.y0
                    obj.x0 += delta_x0
                    obj.y0 += delta_y0

                    # pylint: disable=unused-argument
                    def translate_coords(obj, orig, coords):
                        """Apply translation to coords"""
                        coords[:, ::2] += delta_x0
                        coords[:, 1::2] += delta_y0

                    obj.transform_shapes(None, translate_coords)
                if param.direction == "row":
                    # Distributing images over rows
                    sign = np.sign(param.rows)
                    g_row = (g_row + sign) % param.rows
                    y0 += (obj.height + param.rowspac) * sign
                    if g_row == 0:
                        g_col += 1
                        x0 += obj.width + param.colspac
                        y0 = y0_0
                else:
                    # Distributing images over columns
                    sign = np.sign(param.cols)
                    g_col = (g_col + sign) % param.cols
                    x0 += (obj.width + param.colspac) * sign
                    if g_col == 0:
                        g_row += 1
                        x0 = x0_0
                        y0 += obj.height + param.rowspac
        self.panel.SIG_REFRESH_PLOT.emit("selected", True)

    @qt_try_except()
    def reset_positions(self) -> None:
        """Reset image positions"""
        x0_0, y0_0 = 0.0, 0.0
        delta_x0, delta_y0 = 0.0, 0.0
        objs = self.panel.objview.get_sel_objects(include_groups=True)
        for i_row, obj in enumerate(objs):
            if i_row == 0:
                x0_0, y0_0 = obj.x0, obj.y0
            else:
                delta_x0, delta_y0 = x0_0 - obj.x0, y0_0 - obj.y0
                obj.x0 += delta_x0
                obj.y0 += delta_y0

                # pylint: disable=unused-argument
                def translate_coords(obj, orig, coords):
                    """Apply translation to coords"""
                    coords[:, ::2] += delta_x0
                    coords[:, 1::2] += delta_y0

                obj.transform_shapes(None, translate_coords)
        self.panel.SIG_REFRESH_PLOT.emit("selected", True)

    @qt_try_except()
    def compute_resize(self, param: cdl.param.ResizeParam | None = None) -> None:
        """Resize image with :py:func:`cdl.computation.image.compute_resize`"""
        obj0 = self.panel.objview.get_sel_objects(include_groups=True)[0]
        for obj in self.panel.objview.get_sel_objects():
            if obj.data.shape != obj0.data.shape:
                QW.QMessageBox.warning(
                    self.panel.parent(),
                    APP_NAME,
                    _("Warning:")
                    + "\n"
                    + _("Selected images do not have the same size"),
                )
        edit, param = self.init_param(param, cpi.ResizeParam, _("Resize"))
        if edit:
            original_size = obj0.data.shape
            dlg = ResizeDialog(
                self.plotwidget,
                new_size=original_size,
                old_size=original_size,
                text=_("Destination size:"),
            )
            if not exec_dialog(dlg):
                return
            param.zoom = dlg.get_zoom()
        self.compute_11(cpi.compute_resize, param, title=_("Resize"), edit=edit)

    @qt_try_except()
    def compute_binning(self, param: cdl.param.BinningParam | None = None) -> None:
        """Binning image with :py:func:`cdl.computation.image.compute_binning`"""
        edit = param is None
        obj0 = self.panel.objview.get_sel_objects(include_groups=True)[0]
        input_dtype_str = str(obj0.data.dtype)
        title = _("Binning")
        edit, param = self.init_param(param, cpi.BinningParam, title)
        if edit:
            param.dtype_str = input_dtype_str
        if param.dtype_str is None:
            param.dtype_str = input_dtype_str
        self.compute_11(cpi.compute_binning, param, title=title, edit=edit)

    @qt_try_except()
    def compute_line_profile(
        self, param: cdl.param.LineProfileParam | None = None
    ) -> None:
        """Compute profile along a vertical or horizontal line
        with :py:func:`cdl.computation.image.compute_line_profile`"""
        title = _("Profile")
        add_initial_shape = self.has_param_defaults(cdl.param.LineProfileParam)
        edit, param = self.init_param(param, cpi.LineProfileParam, title)
        if edit:
            options = self.panel.plothandler.get_current_plot_options()
            dlg = ProfileExtractionDialog(
                "line", param, options, self.panel.parent(), add_initial_shape
            )
            obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
            dlg.set_obj(obj)
            if not exec_dialog(dlg):
                return
        self.compute_11(cpi.compute_line_profile, param, title=title, edit=False)

    @qt_try_except()
    def compute_segment_profile(
        self, param: cdl.param.SegmentProfileParam | None = None
    ):
        """Compute profile along a segment
        with :py:func:`cdl.computation.image.compute_segment_profile`"""
        title = _("Profile")
        add_initial_shape = self.has_param_defaults(cdl.param.SegmentProfileParam)
        edit, param = self.init_param(param, cpi.SegmentProfileParam, title)
        if edit:
            options = self.panel.plothandler.get_current_plot_options()
            dlg = ProfileExtractionDialog(
                "segment", param, options, self.panel.parent(), add_initial_shape
            )
            obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
            dlg.set_obj(obj)
            if not exec_dialog(dlg):
                return
        self.compute_11(cpi.compute_segment_profile, param, title=title, edit=False)

    @qt_try_except()
    def compute_average_profile(
        self, param: cdl.param.AverageProfileParam | None = None
    ) -> None:
        """Compute average profile
        with :py:func:`cdl.computation.image.compute_average_profile`"""
        title = _("Average profile")
        add_initial_shape = self.has_param_defaults(cdl.param.AverageProfileParam)
        edit, param = self.init_param(param, cpi.AverageProfileParam, title)
        if edit:
            options = self.panel.plothandler.get_current_plot_options()
            dlg = ProfileExtractionDialog(
                "rectangle", param, options, self.panel.parent(), add_initial_shape
            )
            obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
            dlg.set_obj(obj)
            if not exec_dialog(dlg):
                return
        self.compute_11(cpi.compute_average_profile, param, title=title, edit=False)

    @qt_try_except()
    def compute_radial_profile(
        self, param: cdl.param.RadialProfileParam | None = None
    ) -> None:
        """Compute radial profile
        with :py:func:`cdl.computation.image.compute_radial_profile`"""
        title = _("Radial profile")
        edit, param = self.init_param(param, cpi.RadialProfileParam, title)
        if edit:
            obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
            param.update_from_image(obj)
        self.compute_11(cpi.compute_radial_profile, param, title=title, edit=edit)

    @qt_try_except()
    def compute_histogram(self, param: cdl.param.HistogramParam | None = None) -> None:
        """Compute histogram with :py:func:`cdl.computation.image.compute_histogram`"""
        self.compute_11(
            cpi.compute_histogram, param, cpi.HistogramParam, title=_("Histogram")
        )

    @qt_try_except()
    def compute_swap_axes(self) -> None:
        """Swap data axes with :py:func:`cdl.computation.image.compute_swap_axes`"""
        self.compute_11(cpi.compute_swap_axes, title=_("Swap axes"))

    @qt_try_except()
    def compute_abs(self) -> None:
        """Compute absolute value with :py:func:`cdl.computation.image.compute_abs`"""
        self.compute_11(cpi.compute_abs, title=_("Absolute value"))

    @qt_try_except()
    def compute_re(self) -> None:
        """Compute real part with :py:func:`cdl.computation.image.compute_re`"""
        self.compute_11(cpi.compute_re, title=_("Real part"))

    @qt_try_except()
    def compute_im(self) -> None:
        """Compute imaginary part with :py:func:`cdl.computation.image.compute_im`"""
        self.compute_11(cpi.compute_im, title=_("Imaginary part"))

    @qt_try_except()
    def compute_astype(self, param: cdl.param.DataTypeIParam | None = None) -> None:
        """Convert data type with :py:func:`cdl.computation.image.compute_astype`"""
        self.compute_11(
            cpi.compute_astype, param, cpi.DataTypeIParam, title=_("Convert data type")
        )

    @qt_try_except()
    def compute_log10(self) -> None:
        """Compute Log10 with :py:func:`cdl.computation.image.compute_log10`"""
        self.compute_11(cpi.compute_log10, title="Log10")

    @qt_try_except()
    def compute_exp(self) -> None:
        """Compute Log10 with :py:func:`cdl.computation.image.compute_exp`"""
        self.compute_11(cpi.compute_exp, title=_("Exponential"))

    @qt_try_except()
    def compute_arithmetic(
        self, obj2: ImageObj | None = None, param: cpb.ArithmeticParam | None = None
    ) -> None:
        """Compute arithmetic operation between two images
        with :py:func:`cdl.computation.image.compute_arithmetic`"""
        self.compute_n1n(
            obj2,
            _("image to operate with"),
            cpi.compute_arithmetic,
            param=param,
            paramclass=cpb.ArithmeticParam,
            title=_("Arithmetic"),
        )

    @qt_try_except()
    def compute_difference(self, obj2: ImageObj | list[ImageObj] | None = None) -> None:
        """Compute difference between two images
        with :py:func:`cdl.computation.image.compute_difference`"""
        self.compute_n1n(
            obj2,
            _("image to subtract"),
            cpi.compute_difference,
            title=_("Difference"),
        )

    @qt_try_except()
    def compute_difference_constant(
        self, param: cpb.ConstantParam | None = None
    ) -> None:
        """Compute difference with a constant
        with :py:func:`cdl.computation.image.compute_difference_constant`"""
        self.compute_11(
            cpi.compute_difference_constant,
            param,
            paramclass=cpb.ConstantParam,
            title=_("Difference with constant"),
        )

    @qt_try_except()
    def compute_quadratic_difference(
        self, obj2: ImageObj | list[ImageObj] | None = None
    ) -> None:
        """Compute quadratic difference between two images
        with :py:func:`cdl.computation.image.compute_quadratic_difference`"""
        self.compute_n1n(
            obj2,
            _("image to subtract"),
            cpi.compute_quadratic_difference,
            title=_("Quadratic difference"),
        )

    @qt_try_except()
    def compute_division(self, obj2: ImageObj | list[ImageObj] | None = None) -> None:
        """Compute division between two images
        with :py:func:`cdl.computation.image.compute_division`"""
        self.compute_n1n(
            obj2,
            _("divider"),
            cpi.compute_division,
            title=_("Division"),
        )

    @qt_try_except()
    def compute_division_constant(self, param: cpb.ConstantParam | None = None) -> None:
        """Compute division by a constant
        with :py:func:`cdl.computation.image.compute_division_constant`"""
        self.compute_11(
            cpi.compute_division_constant,
            param,
            paramclass=cpb.ConstantParam,
            title=_("Division by constant"),
        )

    @qt_try_except()
    def compute_flatfield(
        self,
        obj2: ImageObj | None = None,
        param: cdl.param.FlatFieldParam | None = None,
    ) -> None:
        """Compute flat field correction
        with :py:func:`cdl.computation.image.compute_flatfield`"""
        edit, param = self.init_param(param, cpi.FlatFieldParam, _("Flat field"))
        if edit:
            obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
            param.set_from_datatype(obj.data.dtype)
        self.compute_n1n(
            obj2,
            _("flat field image"),
            cpi.compute_flatfield,
            param=param,
            title=_("Flat field correction"),
            edit=edit,
        )

    # ------Image Processing
    @qt_try_except()
    def compute_calibration(
        self, param: cdl.param.ZCalibrateParam | None = None
    ) -> None:
        """Compute data linear calibration
        with :py:func:`cdl.computation.image.compute_calibration`"""
        self.compute_11(
            cpi.compute_calibration,
            param,
            cpi.ZCalibrateParam,
            _("Linear calibration"),
            "y = a.x + b",
        )

    @qt_try_except()
    def compute_clip(self, param: cpb.ClipParam | None = None) -> None:
        """Compute maximum data clipping
        with :py:func:`cdl.computation.image.compute_clip`"""
        self.compute_11(
            cpi.compute_clip,
            param,
            cpb.ClipParam,
            _("Clipping"),
        )

    @qt_try_except()
    def compute_offset_correction(self, param: ROI2DParam | None = None) -> None:
        """Compute offset correction
        with :py:func:`cdl.computation.image.compute_offset_correction`"""
        obj = self.panel.objview.get_sel_objects(include_groups=True)[0]
        if param is None:
            dlg = imagebackground.ImageBackgroundDialog(obj, parent=self.panel.parent())
            if exec_dialog(dlg):
                param = ROI2DParam.create(geometry="rectangle")
                x0, y0, x1, y1 = dlg.get_rect_coords()
                param.x0, param.y0, param.dx, param.dy = x0, y0, x1 - x0, y1 - y0
            else:
                return
        self.compute_11(cpi.compute_offset_correction, param)

    @qt_try_except()
    def compute_gaussian_filter(self, param: cpb.GaussianParam | None = None) -> None:
        """Compute gaussian filter
        with :py:func:`cdl.computation.image.compute_gaussian_filter`"""
        self.compute_11(
            cpi.compute_gaussian_filter, param, cpb.GaussianParam, _("Gaussian filter")
        )

    @qt_try_except()
    def compute_moving_average(
        self, param: cpb.MovingAverageParam | None = None
    ) -> None:
        """Compute moving average
        with :py:func:`cdl.computation.image.compute_moving_average`"""
        self.compute_11(
            cpi.compute_moving_average,
            param,
            cpb.MovingAverageParam,
            _("Moving average"),
        )

    @qt_try_except()
    def compute_moving_median(self, param: cpb.MovingMedianParam | None = None) -> None:
        """Compute moving median
        with :py:func:`cdl.computation.image.compute_moving_median`"""
        self.compute_11(
            cpi.compute_moving_median,
            param,
            cpb.MovingMedianParam,
            _("Moving median"),
        )

    @qt_try_except()
    def compute_wiener(self) -> None:
        """Compute Wiener filter
        with :py:func:`cdl.computation.image.compute_wiener`"""
        self.compute_11(cpi.compute_wiener, title=_("Wiener filter"))

    @qt_try_except()
    def compute_fft(self, param: cdl.param.FFTParam | None = None) -> None:
        """Compute FFT with :py:func:`cdl.computation.image.compute_fft`"""
        if param is None:
            param = cpb.FFTParam.create(shift=Conf.proc.fft_shift_enabled.get())
        self.compute_11(cpi.compute_fft, param, title="FFT", edit=False)

    @qt_try_except()
    def compute_ifft(self, param: cdl.param.FFTParam | None = None) -> None:
        """Compute iFFT with :py:func:`cdl.computation.image.compute_ifft`"""
        if param is None:
            param = cpb.FFTParam.create(shift=Conf.proc.fft_shift_enabled.get())
        self.compute_11(cpi.compute_ifft, param, title="iFFT", edit=False)

    @qt_try_except()
    def compute_magnitude_spectrum(
        self, param: cdl.param.SpectrumParam | None = None
    ) -> None:
        """Compute magnitude spectrum
        with :py:func:`cdl.computation.image.compute_magnitude_spectrum`"""
        self.compute_11(
            cpi.compute_magnitude_spectrum,
            param,
            cpi.SpectrumParam,
            _("Magnitude spectrum"),
        )

    @qt_try_except()
    def compute_phase_spectrum(self) -> None:
        """Compute phase spectrum
        with :py:func:`cdl.computation.image.compute_phase_spectrum`"""
        self.compute_11(cpi.compute_phase_spectrum, title="Phase spectrum")

    @qt_try_except()
    def compute_psd(self, param: cdl.param.SpectrumParam | None = None) -> None:
        """Compute Power Spectral Density (PSD)
        with :py:func:`cdl.computation.image.compute_psd`"""
        self.compute_11(cpi.compute_psd, param, cpi.SpectrumParam, _("PSD"))

    @qt_try_except()
    def compute_butterworth(
        self, param: cdl.param.ButterworthParam | None = None
    ) -> None:
        """Compute Butterworth filter
        with :py:func:`cdl.computation.image.compute_butterworth`"""
        self.compute_11(
            cpi.compute_butterworth,
            param,
            cpi.ButterworthParam,
            _("Butterworth filter"),
        )

    @qt_try_except()
    def compute_threshold(self, param: cdl.param.ThresholdParam | None = None) -> None:
        """Compute parametric threshold
        with :py:func:`cdl.computation.image.threshold.compute_threshold`"""
        self.compute_11(
            cpi_thr.compute_threshold,
            param,
            cpi_thr.ThresholdParam,
            _("Parametric threshold"),
        )

    @qt_try_except()
    def compute_threshold_isodata(self) -> None:
        """Compute threshold using Isodata algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_isodata`"""
        self.compute_11(cpi_thr.compute_threshold_isodata, title="ISODATA")

    @qt_try_except()
    def compute_threshold_li(self) -> None:
        """Compute threshold using Li algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_li`"""
        self.compute_11(cpi_thr.compute_threshold_li, title="Li")

    @qt_try_except()
    def compute_threshold_mean(self) -> None:
        """Compute threshold using Mean algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_mean`"""
        self.compute_11(cpi_thr.compute_threshold_mean, title=_("Mean"))

    @qt_try_except()
    def compute_threshold_minimum(self) -> None:
        """Compute threshold using Minimum algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_minimum`"""
        self.compute_11(cpi_thr.compute_threshold_minimum, title=_("Minimum"))

    @qt_try_except()
    def compute_threshold_otsu(self) -> None:
        """Compute threshold using Otsu algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_otsu`"""
        self.compute_11(cpi_thr.compute_threshold_otsu, title="Otsu")

    @qt_try_except()
    def compute_threshold_triangle(self) -> None:
        """Compute threshold using Triangle algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_triangle`"""
        self.compute_11(cpi_thr.compute_threshold_triangle, title=_("Triangle"))

    @qt_try_except()
    def compute_threshold_yen(self) -> None:
        """Compute threshold using Yen algorithm
        with :py:func:`cdl.computation.image.threshold.compute_threshold_yen`"""
        self.compute_11(cpi_thr.compute_threshold_yen, title="Yen")

    @qt_try_except()
    def compute_all_threshold(self) -> None:
        """Compute all threshold algorithms
        using the following functions:

        - :py:func:`cdl.computation.image.threshold.compute_threshold_isodata`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_li`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_mean`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_minimum`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_otsu`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_triangle`
        - :py:func:`cdl.computation.image.threshold.compute_threshold_yen`
        """
        self.compute_1n(
            [
                cpi_thr.compute_threshold_isodata,
                cpi_thr.compute_threshold_li,
                cpi_thr.compute_threshold_mean,
                cpi_thr.compute_threshold_minimum,
                cpi_thr.compute_threshold_otsu,
                cpi_thr.compute_threshold_triangle,
                cpi_thr.compute_threshold_yen,
            ],
            None,
            "Threshold",
            edit=False,
        )

    @qt_try_except()
    def compute_adjust_gamma(
        self, param: cdl.param.AdjustGammaParam | None = None
    ) -> None:
        """Compute gamma correction
        with :py:func:`cdl.computation.image.exposure.compute_adjust_gamma`"""
        self.compute_11(
            cpi_exp.compute_adjust_gamma,
            param,
            cpi_exp.AdjustGammaParam,
            _("Gamma correction"),
        )

    @qt_try_except()
    def compute_adjust_log(self, param: cdl.param.AdjustLogParam | None = None) -> None:
        """Compute log correction
        with :py:func:`cdl.computation.image.exposure.compute_adjust_log`"""
        self.compute_11(
            cpi_exp.compute_adjust_log,
            param,
            cpi_exp.AdjustLogParam,
            _("Log correction"),
        )

    @qt_try_except()
    def compute_adjust_sigmoid(
        self,
        param: cdl.param.AdjustSigmoidParam | None = None,
    ) -> None:
        """Compute sigmoid correction
        with :py:func:`cdl.computation.image.exposure.compute_adjust_sigmoid`"""
        self.compute_11(
            cpi_exp.compute_adjust_sigmoid,
            param,
            cpi_exp.AdjustSigmoidParam,
            _("Sigmoid correction"),
        )

    @qt_try_except()
    def compute_rescale_intensity(
        self,
        param: cdl.param.RescaleIntensityParam | None = None,
    ) -> None:
        """Rescale image intensity levels
        with :py:func`cdl.computation.image.exposure.compute_rescale_intensity`"""
        self.compute_11(
            cpi_exp.compute_rescale_intensity,
            param,
            cpi_exp.RescaleIntensityParam,
            _("Rescale intensity"),
        )

    @qt_try_except()
    def compute_equalize_hist(
        self, param: cdl.param.EqualizeHistParam | None = None
    ) -> None:
        """Histogram equalization
        with :py:func:`cdl.computation.image.exposure.compute_equalize_hist`"""
        self.compute_11(
            cpi_exp.compute_equalize_hist,
            param,
            cpi_exp.EqualizeHistParam,
            _("Histogram equalization"),
        )

    @qt_try_except()
    def compute_equalize_adapthist(
        self,
        param: cdl.param.EqualizeAdaptHistParam | None = None,
    ) -> None:
        """Adaptive histogram equalization
        with :py:func:`cdl.computation.image.exposure.compute_equalize_adapthist`"""
        self.compute_11(
            cpi_exp.compute_equalize_adapthist,
            param,
            cpi_exp.EqualizeAdaptHistParam,
            _("Adaptive histogram equalization"),
        )

    @qt_try_except()
    def compute_denoise_tv(self, param: cdl.param.DenoiseTVParam | None = None) -> None:
        """Compute Total Variation denoising
        with :py:func:`cdl.computation.image.restoration.compute_denoise_tv`"""
        self.compute_11(
            cpi_res.compute_denoise_tv,
            param,
            cpi_res.DenoiseTVParam,
            _("Total variation denoising"),
        )

    @qt_try_except()
    def compute_denoise_bilateral(
        self,
        param: cdl.param.DenoiseBilateralParam | None = None,
    ) -> None:
        """Compute bilateral filter denoising
        with :py:func:`cdl.computation.image.restoration.compute_denoise_bilateral`"""
        self.compute_11(
            cpi_res.compute_denoise_bilateral,
            param,
            cpi_res.DenoiseBilateralParam,
            _("Bilateral filter denoising"),
        )

    @qt_try_except()
    def compute_denoise_wavelet(
        self,
        param: cdl.param.DenoiseWaveletParam | None = None,
    ) -> None:
        """Compute Wavelet denoising
        with :py:func:`cdl.computation.image.restoration.compute_denoise_wavelet`"""
        self.compute_11(
            cpi_res.compute_denoise_wavelet,
            param,
            cpi_res.DenoiseWaveletParam,
            _("Wavelet denoising"),
        )

    @qt_try_except()
    def compute_denoise_tophat(
        self, param: cdl.param.MorphologyParam | None = None
    ) -> None:
        """Denoise using White Top-Hat
        with :py:func:`cdl.computation.image.restoration.compute_denoise_tophat`"""
        self.compute_11(
            cpi_res.compute_denoise_tophat,
            param,
            cpi_mor.MorphologyParam,
            _("Denoise / Top-Hat"),
        )

    @qt_try_except()
    def compute_all_denoise(self, params: list | None = None) -> None:
        """Compute all denoising filters
        using the following functions:

        - :py:func:`cdl.computation.image.restoration.compute_denoise_tv`
        - :py:func:`cdl.computation.image.restoration.compute_denoise_bilateral`
        - :py:func:`cdl.computation.image.restoration.compute_denoise_wavelet`
        - :py:func:`cdl.computation.image.restoration.compute_denoise_tophat`
        """
        if params is not None:
            assert len(params) == 4, "Wrong number of parameters (4 expected)"
        funcs = [
            cpi_res.compute_denoise_tv,
            cpi_res.compute_denoise_bilateral,
            cpi_res.compute_denoise_wavelet,
            cpi_res.compute_denoise_tophat,
        ]
        edit = params is None
        if edit:
            params = []
            for paramclass, title in (
                (cpi_res.DenoiseTVParam, _("Total variation denoising")),
                (cpi_res.DenoiseBilateralParam, _("Bilateral filter denoising")),
                (cpi_res.DenoiseWaveletParam, _("Wavelet denoising")),
                (cpi_mor.MorphologyParam, _("Denoise / Top-Hat")),
            ):
                param = paramclass(title)
                self.update_param_defaults(param)
                params.append(param)
        self.compute_1n(funcs, params, "Denoise", edit=edit)

    @qt_try_except()
    def compute_white_tophat(
        self, param: cdl.param.MorphologyParam | None = None
    ) -> None:
        """Compute White Top-Hat
        with :py:func:`cdl.computation.image.morphology.compute_white_tophat`"""
        self.compute_11(
            cpi_mor.compute_white_tophat,
            param,
            cpi_mor.MorphologyParam,
            _("White Top-Hat"),
        )

    @qt_try_except()
    def compute_black_tophat(
        self, param: cdl.param.MorphologyParam | None = None
    ) -> None:
        """Compute Black Top-Hat
        with :py:func:`cdl.computation.image.morphology.compute_black_tophat`"""
        self.compute_11(
            cpi_mor.compute_black_tophat,
            param,
            cpi_mor.MorphologyParam,
            _("Black Top-Hat"),
        )

    @qt_try_except()
    def compute_erosion(self, param: cdl.param.MorphologyParam | None = None) -> None:
        """Compute Erosion
        with :py:func:`cdl.computation.image.morphology.compute_erosion`"""
        self.compute_11(
            cpi_mor.compute_erosion,
            param,
            cpi_mor.MorphologyParam,
            _("Erosion"),
        )

    @qt_try_except()
    def compute_dilation(self, param: cdl.param.MorphologyParam | None = None) -> None:
        """Compute Dilation
        with :py:func:`cdl.computation.image.morphology.compute_dilation`"""
        self.compute_11(
            cpi_mor.compute_dilation,
            param,
            cpi_mor.MorphologyParam,
            _("Dilation"),
        )

    @qt_try_except()
    def compute_opening(self, param: cdl.param.MorphologyParam | None = None) -> None:
        """Compute morphological opening
        with :py:func:`cdl.computation.image.morphology.compute_opening`"""
        self.compute_11(
            cpi_mor.compute_opening,
            param,
            cpi_mor.MorphologyParam,
            _("Opening"),
        )

    @qt_try_except()
    def compute_closing(self, param: cdl.param.MorphologyParam | None = None) -> None:
        """Compute morphological closing
        with :py:func:`cdl.computation.image.morphology.compute_closing`"""
        self.compute_11(
            cpi_mor.compute_closing,
            param,
            cpi_mor.MorphologyParam,
            _("Closing"),
        )

    @qt_try_except()
    def compute_all_morphology(
        self, param: cdl.param.MorphologyParam | None = None
    ) -> None:
        """Compute all morphology filters
        using the following functions:

        - :py:func:`cdl.computation.image.morphology.compute_white_tophat`
        - :py:func:`cdl.computation.image.morphology.compute_black_tophat`
        - :py:func:`cdl.computation.image.morphology.compute_erosion`
        - :py:func:`cdl.computation.image.morphology.compute_dilation`
        - :py:func:`cdl.computation.image.morphology.compute_opening`
        - :py:func:`cdl.computation.image.morphology.compute_closing`
        """
        if param is None:
            param = cpi_mor.MorphologyParam()
            if not param.edit(parent=self.panel.parent()):
                return
        funcs = [
            cpi_mor.compute_white_tophat,
            cpi_mor.compute_black_tophat,
            cpi_mor.compute_erosion,
            cpi_mor.compute_dilation,
            cpi_mor.compute_opening,
            cpi_mor.compute_closing,
        ]
        self.compute_1n(funcs, [param] * len(funcs), "Morph", edit=False)

    @qt_try_except()
    def compute_canny(self, param: cdl.param.CannyParam | None = None) -> None:
        """Compute Canny filter
        with :py:func:`cdl.computation.image.edges.compute_canny`"""
        self.compute_11(
            cpi_edg.compute_canny,
            param,
            cpi_edg.CannyParam,
            _("Canny filter"),
        )

    @qt_try_except()
    def compute_roberts(self) -> None:
        """Compute Roberts filter
        with :py:func:`cdl.computation.image.edges.compute_roberts`"""
        self.compute_11(cpi_edg.compute_roberts, title=_("Roberts filter"))

    @qt_try_except()
    def compute_prewitt(self) -> None:
        """Compute Prewitt filter
        with :py:func:`cdl.computation.image.edges.compute_prewitt`"""
        self.compute_11(cpi_edg.compute_prewitt, title=_("Prewitt filter"))

    @qt_try_except()
    def compute_prewitt_h(self) -> None:
        """Compute Prewitt filter (horizontal)
        with :py:func:`cdl.computation.image.edges.compute_prewitt_h`"""
        self.compute_11(
            cpi_edg.compute_prewitt_h,
            title=_("Prewitt filter (horizontal)"),
        )

    @qt_try_except()
    def compute_prewitt_v(self) -> None:
        """Compute Prewitt filter (vertical)
        with :py:func:`cdl.computation.image.edges.compute_prewitt_v`"""
        self.compute_11(
            cpi_edg.compute_prewitt_v,
            title=_("Prewitt filter (vertical)"),
        )

    @qt_try_except()
    def compute_sobel(self) -> None:
        """Compute Sobel filter
        with :py:func:`cdl.computation.image.edges.compute_sobel`"""
        self.compute_11(cpi_edg.compute_sobel, title=_("Sobel filter"))

    @qt_try_except()
    def compute_sobel_h(self) -> None:
        """Compute Sobel filter (horizontal)
        with :py:func:`cdl.computation.image.edges.compute_sobel_h`"""
        self.compute_11(
            cpi_edg.compute_sobel_h,
            title=_("Sobel filter (horizontal)"),
        )

    @qt_try_except()
    def compute_sobel_v(self) -> None:
        """Compute Sobel filter (vertical)
        with :py:func:`cdl.computation.image.edges.compute_sobel_v`"""
        self.compute_11(
            cpi_edg.compute_sobel_v,
            title=_("Sobel filter (vertical)"),
        )

    @qt_try_except()
    def compute_scharr(self) -> None:
        """Compute Scharr filter
        with :py:func:`cdl.computation.image.edges.compute_scharr`"""
        self.compute_11(cpi_edg.compute_scharr, title=_("Scharr filter"))

    @qt_try_except()
    def compute_scharr_h(self) -> None:
        """Compute Scharr filter (horizontal)
        with :py:func:`cdl.computation.image.edges.compute_scharr_h`"""
        self.compute_11(
            cpi_edg.compute_scharr_h,
            title=_("Scharr filter (horizontal)"),
        )

    @qt_try_except()
    def compute_scharr_v(self) -> None:
        """Compute Scharr filter (vertical)
        with :py:func:`cdl.computation.image.edges.compute_scharr_v`"""
        self.compute_11(
            cpi_edg.compute_scharr_v,
            title=_("Scharr filter (vertical)"),
        )

    @qt_try_except()
    def compute_farid(self) -> None:
        """Compute Farid filter
        with :py:func:`cdl.computation.image.edges.compute_farid`"""
        self.compute_11(cpi_edg.compute_farid, title=_("Farid filter"))

    @qt_try_except()
    def compute_farid_h(self) -> None:
        """Compute Farid filter (horizontal)
        with :py:func:`cdl.computation.image.edges.compute_farid_h`"""
        self.compute_11(
            cpi_edg.compute_farid_h,
            title=_("Farid filter (horizontal)"),
        )

    @qt_try_except()
    def compute_farid_v(self) -> None:
        """Compute Farid filter (vertical)
        with :py:func:`cdl.computation.image.edges.compute_farid_v`"""
        self.compute_11(
            cpi_edg.compute_farid_v,
            title=_("Farid filter (vertical)"),
        )

    @qt_try_except()
    def compute_laplace(self) -> None:
        """Compute Laplace filter
        with :py:func:`cdl.computation.image.edges.compute_laplace`"""
        self.compute_11(cpi_edg.compute_laplace, title=_("Laplace filter"))

    @qt_try_except()
    def compute_all_edges(self) -> None:
        """Compute all edges filters
        using the following functions:

        - :py:func:`cdl.computation.image.edges.compute_roberts`
        - :py:func:`cdl.computation.image.edges.compute_prewitt`
        - :py:func:`cdl.computation.image.edges.compute_prewitt_h`
        - :py:func:`cdl.computation.image.edges.compute_prewitt_v`
        - :py:func:`cdl.computation.image.edges.compute_sobel`
        - :py:func:`cdl.computation.image.edges.compute_sobel_h`
        - :py:func:`cdl.computation.image.edges.compute_sobel_v`
        - :py:func:`cdl.computation.image.edges.compute_scharr`
        - :py:func:`cdl.computation.image.edges.compute_scharr_h`
        - :py:func:`cdl.computation.image.edges.compute_scharr_v`
        - :py:func:`cdl.computation.image.edges.compute_farid`
        - :py:func:`cdl.computation.image.edges.compute_farid_h`
        - :py:func:`cdl.computation.image.edges.compute_farid_v`
        - :py:func:`cdl.computation.image.edges.compute_laplace`
        """
        funcs = [
            cpi_edg.compute_roberts,
            cpi_edg.compute_prewitt,
            cpi_edg.compute_prewitt_h,
            cpi_edg.compute_prewitt_v,
            cpi_edg.compute_sobel,
            cpi_edg.compute_sobel_h,
            cpi_edg.compute_sobel_v,
            cpi_edg.compute_scharr,
            cpi_edg.compute_scharr_h,
            cpi_edg.compute_scharr_v,
            cpi_edg.compute_farid,
            cpi_edg.compute_farid_h,
            cpi_edg.compute_farid_v,
            cpi_edg.compute_laplace,
        ]
        self.compute_1n(funcs, None, "Edges")

    @qt_try_except()
    def _extract_multiple_roi_in_single_object(self, group: gds.DataSetGroup) -> None:
        """Extract multiple Regions Of Interest (ROIs) from data in a single object"""
        self.compute_11(cpi.extract_multiple_roi, group, title=_("Extract ROI"))

    @qt_try_except()
    def _extract_each_roi_in_separate_object(self, group: gds.DataSetGroup) -> None:
        """Extract each single Region Of Interest (ROI) from data in a separate
        object (keep the ROI in the case of a circular ROI, for example)"""
        self.compute_1n(cpi.extract_single_roi, group.datasets, "ROI", edit=False)

    # ------Image Analysis
    @qt_try_except()
    def compute_stats(self) -> dict[str, ResultProperties]:
        """Compute data statistics
        with :py:func:`cdl.computation.image.compute_stats`"""
        return self.compute_10(cpi.compute_stats, title=_("Statistics"))

    @qt_try_except()
    def compute_centroid(self) -> dict[str, ResultShape]:
        """Compute image centroid
        with :py:func:`cdl.computation.image.compute_centroid`"""
        return self.compute_10(cpi.compute_centroid, title=_("Centroid"))

    @qt_try_except()
    def compute_enclosing_circle(self) -> dict[str, ResultShape]:
        """Compute minimum enclosing circle
        with :py:func:`cdl.computation.image.compute_enclosing_circle`"""
        return self.compute_10(
            cpi.compute_enclosing_circle, title=_("Enclosing circle")
        )

    @qt_try_except()
    def compute_peak_detection(
        self, param: cdl.param.Peak2DDetectionParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute 2D peak detection
        with :py:func:`cdl.computation.image.compute_peak_detection`"""
        edit, param = self.init_param(
            param, cpi_det.Peak2DDetectionParam, _("Peak detection")
        )
        if edit:
            data = self.panel.objview.get_sel_objects(include_groups=True)[0].data
            param.size = max(min(data.shape) // 40, 50)

        results = self.compute_10(
            cpi_det.compute_peak_detection,
            param,
            edit=edit,
            title=_("Peak detection"),
        )
        if results is not None and param.create_rois and len(results.items()) > 1:
            with create_progress_bar(
                self.panel, _("Create regions of interest"), max_=len(results)
            ) as progress:
                for idx, (oid, result) in enumerate(results.items()):
                    progress.setValue(idx + 1)
                    QW.QApplication.processEvents()
                    if progress.wasCanceled():
                        break
                    obj = self.panel.objmodel[oid]
                    dist = distance_matrix(result.raw_data)
                    dist_min = dist[dist != 0].min()
                    assert dist_min > 0
                    radius = int(0.5 * dist_min / np.sqrt(2) - 1)
                    assert radius >= 1
                    ymax, xmax = obj.data.shape
                    coords = []
                    for x, y in result.raw_data:
                        x0, y0 = max(x - radius, 0), max(y - radius, 0)
                        dx, dy = min(x + radius, xmax) - x0, min(y + radius, ymax) - y0
                        coords.append([x0, y0, dx, dy])
                    obj.roi = create_image_roi("rectangle", coords, indices=True)
                    self.SIG_ADD_SHAPE.emit(obj.uuid)
                    self.panel.SIG_REFRESH_PLOT.emit(obj.uuid, True)
        return results

    @qt_try_except()
    def compute_contour_shape(
        self, param: cdl.param.ContourShapeParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute contour shape fit
        with :py:func:`cdl.computation.image.detection.compute_contour_shape`"""
        edit, param = self.init_param(param, cpi_det.ContourShapeParam, _("Contour"))
        return self.compute_10(
            cpi_det.compute_contour_shape,
            param=param,
            title=_("Contour"),
            edit=edit,
        )

    @qt_try_except()
    def compute_hough_circle_peaks(
        self, param: cdl.param.HoughCircleParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute peak detection based on a circle Hough transform
        with :py:func:`cdl.computation.image.compute_hough_circle_peaks`"""
        return self.compute_10(
            cpi.compute_hough_circle_peaks,
            param,
            cpi.HoughCircleParam,
            title=_("Hough circles"),
        )

    @qt_try_except()
    def compute_blob_dog(
        self, param: cdl.param.BlobDOGParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute blob detection using Difference of Gaussian method
        with :py:func:`cdl.computation.image.detection.compute_blob_dog`"""
        return self.compute_10(
            cpi_det.compute_blob_dog,
            param,
            cpi_det.BlobDOGParam,
            title=_("Blob detection (DOG)"),
        )

    @qt_try_except()
    def compute_blob_doh(
        self, param: cdl.param.BlobDOHParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute blob detection using Determinant of Hessian method
        with :py:func:`cdl.computation.image.detection.compute_blob_doh`"""
        return self.compute_10(
            cpi_det.compute_blob_doh,
            param,
            cpi_det.BlobDOHParam,
            title=_("Blob detection (DOH)"),
        )

    @qt_try_except()
    def compute_blob_log(
        self, param: cdl.param.BlobLOGParam | None = None
    ) -> dict[str, ResultShape]:
        """Compute blob detection using Laplacian of Gaussian method
        with :py:func:`cdl.computation.image.detection.compute_blob_log`"""
        return self.compute_10(
            cpi_det.compute_blob_log,
            param,
            cpi_det.BlobLOGParam,
            title=_("Blob detection (LOG)"),
        )

    @qt_try_except()
    def compute_blob_opencv(
        self,
        param: cdl.param.BlobOpenCVParam | None = None,
    ) -> dict[str, ResultShape]:
        """Compute blob detection using OpenCV
        with :py:func:`cdl.computation.image.detection.compute_blob_opencv`"""
        return self.compute_10(
            cpi_det.compute_blob_opencv,
            param,
            cpi_det.BlobOpenCVParam,
            title=_("Blob detection (OpenCV)"),
        )



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Create dependencies hash
"""

from cdl.config import DATAPATH
from cdl.utils import dephash

dephash.create_dependencies_file(DATAPATH, ("guidata", "plotpy"))



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Macro Panel unit tests
----------------------

The objective of this test is to check all the functionalities of the Macro Panel
widget, by calling all its methods and checking the results.

Some methods are not tested here, as they are tested in remote control tests
(see cdl/tests/features/control/remoteclient_app.py):
- `run_macro`
- `stop_macro`

All other methods should be tested here.
"""

# guitest: show

import os.path as osp

from guidata.qthelpers import qt_app_context

from cdl.core.gui.macroeditor import Macro
from cdl.core.gui.panel import macro
from cdl.env import execenv
from cdl.utils import tests
from cdl.utils.tests import get_temporary_directory


def get_macro_example_path() -> str:
    """Return macro example path"""
    path = get_temporary_directory()
    contents = """
# Simple DataLab macro example

import numpy as np

from cdl.proxy import RemoteProxy

proxy = RemoteProxy()

z = np.random.rand(20, 20)
proxy.add_image("toto", z)
proxy.compute_fft()

print("All done! :)")
"""
    filename = osp.join(path, "macro_example.py")
    with open(filename, "w", encoding="utf-8") as f:
        f.write(contents)
    return filename


def test_macro_editor():
    """Test dep viewer window"""
    with qt_app_context(exec_loop=True):
        widget = macro.MacroPanel()
        widget.resize(800, 600)
        widget.show()

        # Create a new macro
        new_macro = widget.add_macro()
        assert new_macro is widget.get_macro()

        # Check out the macro title, serializable name, number, ...
        execenv.print("Macro title:", new_macro.title)
        execenv.print("Serializable name:", widget.get_serializable_name(new_macro))
        nb1 = widget.get_number_from_macro(new_macro)
        nb2 = widget.get_number_from_title(new_macro.title)
        assert nb1 == nb2 and nb1 == 1
        execenv.print("Macro number:", nb1)
        titles = widget.get_macro_titles()
        assert titles[0] == new_macro.title and len(titles) == 1
        new_title = "New title"
        widget.rename_macro(1, new_title)
        assert widget.get_macro_titles()[0] == new_title

        with tests.CDLTemporaryDirectory() as tmpdir:
            fname = osp.join(tmpdir, "macro.py")
            widget.export_macro_to_file(1, fname)
            widget.import_macro_from_file(fname)
            imported_macro: Macro = widget.get_macro(2)
            assert imported_macro.title == new_macro.title
            assert imported_macro.get_code() == new_macro.get_code()
            widget.rename_macro(1, "Other title")
            widget.remove_macro(1)
            assert len(widget.get_macro_titles()) == 1
            assert widget.get_macro_titles()[0] == imported_macro.title

        # Remove all macros
        widget.remove_all_objects()
        assert len(widget.get_macro_titles()) == 0

        # Load a macro from file
        macro_path = get_macro_example_path()
        widget.import_macro_from_file(macro_path)
        assert len(widget.get_macro_titles()) == 1
        assert widget.get_macro_titles()[0] == osp.basename(macro_path)


if __name__ == "__main__":
    test_macro_editor()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Arithmetic parameters unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from guidata.qthelpers import qt_app_context

from cdl.env import execenv
from cdl.param import ArithmeticParam


def test_arithmetic_param_interactive():
    """Arithmetic parameters interactive test."""
    with qt_app_context():
        param = ArithmeticParam()
        if param.edit():
            execenv.print(param)


if __name__ == "__main__":
    test_arithmetic_param_interactive()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab plugin system
---------------------

DataLab plugin system provides a way to extend the application with new
functionalities.

Plugins are Python modules that relies on two classes:

    - :class:`PluginInfo`, which stores information about the plugin
    - :class:`PluginBase`, which is the base class for all plugins

Plugins may also extends DataLab I/O features by providing new image or
signal formats. To do so, they must provide a subclass of :class:`ImageFormatBase`
or :class:`SignalFormatBase`, in which format infos are defined using the
:class:`FormatInfo` class.
"""

from __future__ import annotations

import abc
import dataclasses
import importlib
import os
import os.path as osp
import pkgutil
import sys
from typing import TYPE_CHECKING

from qtpy import QtWidgets as QW

from cdl.config import MOD_NAME, OTHER_PLUGINS_PATHLIST, Conf, _

# pylint: disable=unused-import
from cdl.core.io.base import FormatInfo  # noqa: F401
from cdl.core.io.image.base import ImageFormatBase  # noqa: F401
from cdl.core.io.image.formats import ClassicsImageFormat  # noqa: F401
from cdl.core.io.signal.base import SignalFormatBase  # noqa: F401
from cdl.env import execenv
from cdl.proxy import LocalProxy

if TYPE_CHECKING:
    from cdl.core.gui import main
    from cdl.core.gui.panel.image import ImagePanel
    from cdl.core.gui.panel.signal import SignalPanel
    from cdl.core.model.image import NewImageParam
    from cdl.core.model.signal import NewSignalParam


PLUGINS_DEFAULT_PATH = Conf.get_path("plugins")

if not osp.isdir(PLUGINS_DEFAULT_PATH):
    os.makedirs(PLUGINS_DEFAULT_PATH)


#  pylint: disable=bad-mcs-classmethod-argument
class PluginRegistry(type):
    """Metaclass for registering plugins"""

    _plugin_classes: list[type[PluginBase]] = []
    _plugin_instances: list[PluginBase] = []

    def __init__(cls, name, bases, attrs):
        super().__init__(name, bases, attrs)
        if name != "PluginBase":
            cls._plugin_classes.append(cls)

    @classmethod
    def get_plugin_classes(cls) -> list[type[PluginBase]]:
        """Return plugin classes"""
        return cls._plugin_classes

    @classmethod
    def get_plugins(cls) -> list[PluginBase]:
        """Return plugin instances"""
        return cls._plugin_instances

    @classmethod
    def get_plugin(cls, name_or_class: str | type[PluginBase]) -> PluginBase | None:
        """Return plugin instance"""
        for plugin in cls._plugin_instances:
            if name_or_class in (plugin.info.name, plugin.__class__):
                return plugin
        return None

    @classmethod
    def register_plugin(cls, plugin: PluginBase):
        """Register plugin"""
        if plugin.info.name in [plug.info.name for plug in cls._plugin_instances]:
            raise ValueError(f"Plugin {plugin.info.name} already registered")
        cls._plugin_instances.append(plugin)
        execenv.log(cls, f"Plugin {plugin.info.name} registered")

    @classmethod
    def unregister_plugin(cls, plugin: PluginBase):
        """Unregister plugin"""
        cls._plugin_instances.remove(plugin)
        execenv.log(cls, f"Plugin {plugin.info.name} unregistered")
        execenv.log(cls, f"{len(cls._plugin_instances)} plugins left")

    @classmethod
    def get_plugin_infos(cls, html: bool = True) -> str:
        """Return plugin infos (names, versions, descriptions) in html format

        Args:
            html: return html formatted text (default: True)
        """
        linesep = "<br>" if html else os.linesep
        bullet = "• " if html else " " * 4

        def italic(text: str) -> str:
            """Return italic text"""
            return f"<i>{text}</i>" if html else text

        if Conf.main.plugins_enabled.get():
            plugins = cls.get_plugins()
            if plugins:
                text = italic(_("Registered plugins:"))
                text += linesep
                for plugin in plugins:
                    text += f"{bullet}{plugin.info.name} ({plugin.info.version})"
                    if plugin.info.description:
                        text += f": {plugin.info.description}"
                    text += linesep
            else:
                text = italic(_("No plugins available"))
        else:
            text = italic(_("Plugins are disabled (see DataLab settings)"))
        return text


@dataclasses.dataclass
class PluginInfo:
    """Plugin info"""

    name: str = None
    version: str = "0.0.0"
    description: str = ""
    icon: str = None


class PluginBaseMeta(PluginRegistry, abc.ABCMeta):
    """Mixed metaclass to avoid conflicts"""


class PluginBase(abc.ABC, metaclass=PluginBaseMeta):
    """Plugin base class"""

    PLUGIN_INFO: PluginInfo = None

    def __init__(self):
        self.main: main.CDLMainWindow = None
        self.proxy: LocalProxy = None
        self._is_registered = False
        self.info = self.PLUGIN_INFO
        if self.info is None:
            raise ValueError(f"Plugin info not set for {self.__class__.__name__}")

    @property
    def signalpanel(self) -> SignalPanel:
        """Return signal panel"""
        return self.main.signalpanel

    @property
    def imagepanel(self) -> ImagePanel:
        """Return image panel"""
        return self.main.imagepanel

    def show_warning(self, message: str):
        """Show warning message"""
        QW.QMessageBox.warning(self.main, _("Warning"), message)

    def show_error(self, message: str):
        """Show error message"""
        QW.QMessageBox.critical(self.main, _("Error"), message)

    def show_info(self, message: str):
        """Show info message"""
        QW.QMessageBox.information(self.main, _("Information"), message)

    def ask_yesno(
        self, message: str, title: str | None = None, cancelable: bool = False
    ) -> bool:
        """Ask yes/no question"""
        if title is None:
            title = _("Question")
        buttons = QW.QMessageBox.Yes | QW.QMessageBox.No
        if cancelable:
            buttons |= QW.QMessageBox.Cancel
        answer = QW.QMessageBox.question(self.main, title, message, buttons)
        if answer == QW.QMessageBox.Yes:
            return True
        if answer == QW.QMessageBox.No:
            return False
        return None

    def edit_new_signal_parameters(
        self,
        title: str | None = None,
        size: int | None = None,
        hide_signal_type: bool = True,
    ) -> NewSignalParam:
        """Create and edit new signal parameter dataset

        Args:
            title: title of the new signal
            size: size of the new signal (default: None, get from current signal)
            hide_signal_type: hide signal type parameter (default: True)

        Returns:
            New signal parameter dataset (or None if canceled)
        """
        newparam = self.signalpanel.get_newparam_from_current(title=title)
        if size is not None:
            newparam.size = size
        newparam.hide_signal_type = hide_signal_type
        if newparam.edit(self.main):
            return newparam
        return None

    def edit_new_image_parameters(
        self,
        title: str | None = None,
        shape: tuple[int, int] | None = None,
        hide_image_type: bool = True,
        hide_image_dtype: bool = False,
    ) -> NewImageParam | None:
        """Create and edit new image parameter dataset

        Args:
            title: title of the new image
            shape: shape of the new image (default: None, get from current image)
            hide_image_type: hide image type parameter (default: True)
            hide_image_dtype: hide image data type parameter (default: False)

        Returns:
            New image parameter dataset (or None if canceled)
        """
        newparam = self.imagepanel.get_newparam_from_current(title=title)
        if shape is not None:
            newparam.width, newparam.height = shape
        newparam.hide_image_type = hide_image_type
        newparam.hide_image_dtype = hide_image_dtype
        if newparam.edit(self.main):
            return newparam
        return None

    def is_registered(self):
        """Return True if plugin is registered"""
        return self._is_registered

    def register(self, main: main.CDLMainWindow) -> None:
        """Register plugin"""
        if self._is_registered:
            return
        PluginRegistry.register_plugin(self)
        self._is_registered = True
        self.main = main
        self.proxy = LocalProxy(main)
        self.register_hooks()

    def unregister(self):
        """Unregister plugin"""
        if not self._is_registered:
            return
        PluginRegistry.unregister_plugin(self)
        self._is_registered = False
        self.unregister_hooks()
        self.main = None
        self.proxy = None

    def register_hooks(self):
        """Register plugin hooks"""

    def unregister_hooks(self):
        """Unregister plugin hooks"""

    @abc.abstractmethod
    def create_actions(self):
        """Create actions"""


def discover_plugins() -> list[type[PluginBase]]:
    """Discover plugins using naming convention

    Returns:
        List of discovered plugins (as classes)
    """
    if Conf.main.plugins_enabled.get():
        for path in [
            Conf.main.plugins_path.get(),
            PLUGINS_DEFAULT_PATH,
        ] + OTHER_PLUGINS_PATHLIST:
            rpath = osp.realpath(path)
            if rpath not in sys.path:
                sys.path.append(rpath)
        return [
            importlib.import_module(name)
            for _finder, name, _ispkg in pkgutil.iter_modules()
            if name.startswith(f"{MOD_NAME}_")
        ]
    return []


def get_available_plugins() -> list[PluginBase]:
    """Instantiate and get available plugins

    Returns:
        List of available plugins (as instances)
    """
    # Note: this function is not used by DataLab itself, but it is used by the
    #       test suite to get a list of available plugins
    discover_plugins()
    return [plugin_class() for plugin_class in PluginRegistry.get_plugin_classes()]



# -*- coding: utf-8 -*-
"""
Example of remote control of DataLab current session,
from a Python script running outside DataLab (e.g. in Spyder)

Created on Fri May 12 12:28:56 2023

@author: p.raybaut
"""

# %% Importing necessary modules

import numpy as np
import scipy.ndimage as spi
from cdlclient import SimpleRemoteProxy

# %% Connecting to DataLab current session

proxy = SimpleRemoteProxy()
proxy.connect()

# %% Executing commands in DataLab (...)


# Define our custom processing function
def weighted_average_denoise(data: np.ndarray) -> np.ndarray:
    """Apply a custom denoising filter to an image.

    This filter averages the pixels in a 5x5 neighborhood, but gives less weight
    to pixels that significantly differ from the central pixel.
    """

    def filter_func(values: np.ndarray) -> float:
        """Filter function"""
        central_pixel = values[len(values) // 2]
        differences = np.abs(values - central_pixel)
        weights = np.exp(-differences / np.mean(differences))
        return np.average(values, weights=weights)

    return spi.generic_filter(data, filter_func, size=5)


# Switch to the "Image Panel" and get the current image
proxy.set_current_panel("image")
image = proxy.get_object()
if image is None:
    # We raise an explicit error if there is no image to process
    raise RuntimeError("No image to process!")

# Get a copy of the image data, and apply the function to it
data = np.array(image.data, copy=True)
data = weighted_average_denoise(data)

# Add new image to the panel
proxy.add_image("Filtered using Spyder", data)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Basic application launcher test 1

Running application a few times in a row with different entry parameters.
"""

# guitest: show

from cdl import app
from cdl.env import execenv
from cdl.utils.qthelpers import cdl_app_context


def test_launcher1(screenshots: bool = False):
    """Testing DataLab app launcher"""
    with cdl_app_context(exec_loop=not screenshots):
        execenv.print("Opening DataLab with no argument")
        win = app.create()
        if screenshots:
            win.statusBar().hide()
            win.take_screenshot("s_app_at_startup")
            win.close()


if __name__ == "__main__":
    test_launcher1()



# Simple DataLab macro example

import numpy as np

# When using this code outside of DataLab (which is possible!), you may use the
# *Simple DataLab Client* (see https://pypi.org/project/cdlclient/), installed with
# `pip install cdlclient`), as it's more lightweight than the full DataLab package:
#
# from cdlclient import SimpleRemoteProxy as RemoteProxy
from cdl.proxy import RemoteProxy

proxy = RemoteProxy()

z = np.random.rand(20, 20)
proxy.add_image("toto", z)
proxy.compute_fft()

print("All done!")



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Statistics unit test

Testing the following:
  - Create a signal
  - Compute statistics on signal and compare with expected results
  - Create an image
  - Compute statistics on image and compare with expected results
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

import numpy as np
import pytest
import scipy.integrate as spt

import cdl.computation.image as cpi
import cdl.computation.signal as cps
import cdl.obj


def get_analytical_stats(data: np.ndarray) -> dict[str, float]:
    """Compute analytical statistics for data

    Args:
        data: Array of data

    Returns:
        Dictionary with analytical statistics
    """
    results = {}
    if data.shape[0] == 2:
        # This is a signal data (row 0: x, row 1: y)
        results["trapz"] = spt.trapezoid(data[1], data[0])
        data = data[1]
    results.update(
        {
            "min": np.min(data),
            "max": np.max(data),
            "mean": np.mean(data),
            "median": np.median(data),
            "std": np.std(data),
            "mean/std": np.mean(data) / np.std(data),
            "ptp": np.ptp(data),
            "sum": np.sum(data),
        }
    )
    return results


def create_reference_signal() -> cdl.obj.SignalObj:
    """Create reference signal"""
    snew = cdl.obj.new_signal_param("Gaussian", stype=cdl.obj.SignalTypes.GAUSS)
    addparam = cdl.obj.GaussLorentzVoigtParam()
    sig = cdl.obj.create_signal_from_param(snew, addparam=addparam, edit=False)
    sig.roi = cdl.obj.create_signal_roi([len(sig.x) // 2, len(sig.x) - 1], indices=True)
    return sig


def create_reference_image() -> cdl.obj.ImageObj:
    """Create reference image"""
    inew = cdl.obj.new_image_param("2D-Gaussian", cdl.obj.ImageTypes.GAUSS)
    addparam = cdl.obj.Gauss2DParam()
    ima = cdl.obj.create_image_from_param(inew, addparam=addparam, edit=False)
    dy, dx = ima.data.shape
    ima.roi = cdl.obj.create_image_roi(
        "rectangle",
        [
            [dx // 2, 0, dx, dy],
            [0, 0, dx // 3, dy // 3],
            [dx // 2, dy // 2, dx, dy],
        ],
    )
    return ima


@pytest.mark.validation
def test_signal_stats_unit() -> None:
    """Validate computed statistics for signals"""
    obj = create_reference_signal()
    res = cps.compute_stats(obj)
    df = res.to_dataframe()
    ref = get_analytical_stats(obj.xydata)
    name_map = {
        "min": "min(y)",
        "max": "max(y)",
        "mean": "<y>",
        "median": "median(y)",
        "std": "σ(y)",
        "mean/std": "<y>/σ(y)",
        "ptp": "peak-to-peak(y)",
        "sum": "Σ(y)",
        "trapz": "∫ydx",
    }
    for key, val in ref.items():
        colname = name_map[key]
        assert colname in df
        assert np.isclose(df[colname][0], val), f"Incorrect value for {colname}"

    # Given the fact that signal ROI is set to [len(sig.x) // 2, len(sig.x) - 1],
    # we may check the relationship between the results on the whole signal and the ROI:
    for key, val in ref.items():
        colname = name_map[key]
        if key in ("trapz", "sum"):
            assert np.isclose(df[colname][1], val / 2, rtol=0.02)
        elif key == "median":
            continue
        else:
            assert np.isclose(df[colname][1], val, rtol=0.01)


@pytest.mark.validation
def test_image_stats_unit() -> None:
    """Validate computed statistics for images"""
    obj = create_reference_image()

    # Ignore "RuntimeWarning: invalid value encountered in scalar divide" in the test
    # (this warning is due to the fact that the 2nd ROI has zero sum of pixel values,
    # hence the mean/std is NaN)
    with np.errstate(invalid="ignore"):
        res = cpi.compute_stats(obj)

    df = res.to_dataframe()
    ref = get_analytical_stats(obj.data)
    name_map = {
        "min": "min(z)",
        "max": "max(z)",
        "mean": "<z>",
        "median": "median(z)",
        "std": "σ(z)",
        "mean/std": "<z>/σ(z)",
        "ptp": "peak-to-peak(z)",
        "sum": "Σ(z)",
    }
    for key, val in ref.items():
        colname = name_map[key]
        assert colname in df
        assert np.isclose(
            df[colname][0], val, rtol=1e-4, atol=1e-5
        ), f"Incorrect value for {colname}"

    # Given the fact that image ROI is set to
    # [[dx // 2, 0, dx, dy], [0, 0, dx // 3, dy // 3], [dx // 2, dy // 2, dx, dy]],
    # we may check the relationship between the results on the whole image and the ROIs:
    for key, val in ref.items():
        colname = name_map[key]
        if key == "sum":
            assert np.isclose(df[colname][1], val / 2, rtol=0.02)
            assert np.isclose(df[colname][3], val / 4, rtol=0.02)
        elif key == "median":
            continue
        else:
            assert np.isclose(df[colname][1], val, rtol=0.01)
            assert np.isclose(df[colname][3], val, rtol=0.01)
            if key != "mean/std":
                assert np.isclose(df[colname][2], 0.0)


if __name__ == "__main__":
    test_signal_stats_unit()
    test_image_stats_unit()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
DataLab Wizard Widget
---------------------

The DataLab Wizard is a widget that guides the user through a series of steps
to complete a task. It is implemented as a series of pages, each of which is
a separate widget.

The `Wizard` class is the main widget that contains the pages. The `WizardPage`
class is the base class for the pages.

This module is strongly inspired from Qt's `QWizard` and `QWizardPage` classes.

.. note::

    The only motivation for reimplementing the wizard widget is to
    support complete styling with `QPalette` and `QStyle` (e.g. `guidata`'s
    dark mode is not supported on Windows).
"""

from __future__ import annotations

from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.compat import getopenfilename
from qtpy.QtWidgets import QWidget

from cdl.config import _


class WizardPage(QW.QWidget):
    """Wizard page base class

    We create our own wizard page class instead of using QWizardPage because
    the latter does not support complete styling with `QPalette` and `QStyle`
    (e.g. `guidata`'s dark mode is not supported on Windows).

    This class reimplements the `QWizardPage` features.

    """

    SIG_INITIALIZE_PAGE = QC.Signal()
    SIG_VALID_STATE_CHANGED = QC.Signal()

    def __init__(self, parent: QW.QWidget | None = None) -> None:
        super().__init__(parent)
        self.__is_valid: bool = True
        self.wizard: Wizard | None = None
        self._main_layout = QW.QVBoxLayout()
        self._user_layout = QW.QVBoxLayout()
        self._title_label = QW.QLabel("")
        font = self._title_label.font()
        font.setPointSize(font.pointSize() + 4)
        font.setBold(True)
        self._title_label.setFont(font)
        self._title_label.setStyleSheet("color: #1E90FF")
        horiz_line = QW.QFrame()
        horiz_line.setFrameShape(QW.QFrame.HLine)
        horiz_line.setFrameShadow(QW.QFrame.Sunken)
        self._subtitle_label = QW.QLabel("")
        self._main_layout.addWidget(self._title_label)
        self._main_layout.addWidget(self._subtitle_label)
        self._main_layout.addWidget(horiz_line)
        self._main_layout.addLayout(self._user_layout)
        self.setLayout(self._main_layout)

    def set_wizard(self, wizard: Wizard) -> None:
        """Set the wizard"""
        self.wizard = wizard

    def get_wizard(self) -> Wizard:
        """Return the wizard"""
        return self.wizard

    def set_title(self, title: str) -> None:
        """Set the title of the page"""
        self._title_label.setText(title)

    def set_subtitle(self, subtitle: str) -> None:
        """Set the subtitle of the page"""
        self._subtitle_label.setText(subtitle)

    def set_valid(self, is_valid: bool) -> None:
        """Set the page as valid"""
        self.__is_valid = is_valid
        self.SIG_VALID_STATE_CHANGED.emit()

    def is_valid(self) -> bool:
        """Return whether the page is valid"""
        return self.__is_valid

    def add_to_layout(self, layout: QW.QLayout | QW.QWidget) -> None:
        """Add a layout to the user layout"""
        if isinstance(layout, QW.QWidget):
            self._user_layout.addWidget(layout)
        else:
            self._user_layout.addLayout(layout)

    def add_stretch(self) -> None:
        """Add a stretch to the user layout"""
        self._user_layout.addStretch()

    def initialize_page(self) -> None:
        """Initialize the page"""
        self.SIG_INITIALIZE_PAGE.emit()

    def validate_page(self) -> bool:
        """Validate the page"""
        return self.is_valid()


class Wizard(QW.QDialog):
    """Wizard base class"""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        _main_layout = QW.QVBoxLayout()
        self.setLayout(_main_layout)

        self._pages_widget = QW.QStackedWidget()
        _main_layout.addWidget(self._pages_widget)

        btn_layout = QW.QHBoxLayout()
        self._back_btn = QW.QPushButton(_("Back"))
        self._back_btn.clicked.connect(self.go_to_previous_page)
        self._next_btn = QW.QPushButton(_("Next"))
        self._next_btn.clicked.connect(self.go_to_next_page)
        self._finish_btn = QW.QPushButton(_("Finish"))
        self._finish_btn.clicked.connect(self.accept)
        self._cancel_btn = QW.QPushButton(_("Cancel"))
        self._cancel_btn.clicked.connect(self.reject)
        btn_layout.addWidget(self._back_btn)
        btn_layout.addWidget(self._next_btn)
        btn_layout.addWidget(self._finish_btn)
        btn_layout.addWidget(self._cancel_btn)
        _main_layout.addLayout(btn_layout)

        self.setSizePolicy(
            QW.QSizePolicy(QW.QSizePolicy.Minimum, QW.QSizePolicy.Minimum)
        )

    def add_page(self, page: WizardPage, last_page: bool = False) -> None:
        """Add a page to the wizard"""
        page.set_wizard(self)
        page.SIG_INITIALIZE_PAGE.connect(self.__update_button_states)
        page.SIG_VALID_STATE_CHANGED.connect(self.__update_button_states)
        self._pages_widget.addWidget(page)
        if last_page:
            self._pages_widget.widget(0).initialize_page()

    def __update_button_states(self, index: int | None = None) -> None:
        """Update button states"""
        if index is None:
            index = self._pages_widget.currentIndex()
        self._back_btn.setEnabled(index > 0)
        not_last_page = index < self._pages_widget.count() - 1
        page_valid = self._pages_widget.currentWidget().is_valid()
        self._next_btn.setEnabled(not_last_page and page_valid)
        is_last_page = index == self._pages_widget.count() - 1
        self._finish_btn.setEnabled(is_last_page and page_valid)

    def go_to_previous_page(self) -> None:
        """Go to the previous page"""
        self._pages_widget.setCurrentIndex(self._pages_widget.currentIndex() - 1)
        self.__update_button_states()

    def go_to_next_page(self) -> None:
        """Go to the next page"""
        if self.validate_page():
            self._pages_widget.setCurrentIndex(self._pages_widget.currentIndex() + 1)
            self.initialize_page()

    def initialize_page(self) -> None:
        """Initialize the page"""
        self._pages_widget.currentWidget().initialize_page()

    def validate_page(self) -> bool:
        """Validate the page"""
        return self._pages_widget.currentWidget().validate_page()

    def accept(self) -> None:
        """Accept the wizard"""
        if self.validate_page():
            super().accept()


class ExamplePage1(WizardPage):
    """Example wizard page 1"""

    def __init__(self) -> None:
        super().__init__()
        self.set_title(_("Welcome to the Example Wizard"))
        self.set_subtitle(
            _("This wizard will guide you through the process of importing data.")
        )

    def initialize_page(self) -> None:
        """Initialize the page"""
        print("ExamplePage1 initialized")
        super().initialize_page()

    def validate_page(self) -> bool:
        """Validate the page"""
        print("ExamplePage1 validated")
        return super().validate_page()


class ExamplePage2(WizardPage):
    """Example wizard page 2"""

    def __init__(self) -> None:
        super().__init__()
        self.set_title(_("Select the Source of the Data"))
        self.set_subtitle(
            _("Select the source of the data to be imported (clipboard or file).")
        )
        self._clipboard_rb = QW.QRadioButton(_("Clipboard"))
        self._file_rb = QW.QRadioButton(_("File"))
        self._file_rb.toggled.connect(self.file_rb_toggled)
        self._file_le = QW.QLineEdit()
        self._file_btn = QW.QPushButton(_("Browse..."))
        self._file_btn.clicked.connect(self.browse_file)
        self.add_to_layout(self._clipboard_rb)
        self.add_to_layout(self._file_rb)
        self.add_to_layout(self._file_le)
        self.add_to_layout(self._file_btn)

    def initialize_page(self) -> None:
        """Initialize the page"""
        print("ExamplePage2 initialized")
        super().initialize_page()

    def file_rb_toggled(self, checked: bool) -> None:
        """File radio button toggled"""
        self._file_le.setEnabled(checked)
        self._file_btn.setEnabled(checked)

    def browse_file(self) -> None:
        """Browse file"""
        file_name, _filt = getopenfilename(
            self,
            _("Select the File to Import"),
            "",
            _("CSV Files (*.csv);;Text Files (*.txt);;All Files (*)"),
        )
        if file_name:
            self._file_le.setText(file_name)

    def validate_page(self) -> bool:
        """Validate the page"""
        if self._file_rb.isChecked() and not self._file_le.text():
            QW.QMessageBox.critical(
                self,
                _("Error"),
                _("Please select the file to import."),
                QW.QMessageBox.Ok,
            )
            return False
        return True


class ExampleWizard(Wizard):
    """Example wizard widget"""

    def __init__(self) -> None:
        super().__init__()
        self.setWindowTitle(_("Example Wizard"))
        self.add_page(ExamplePage1())
        self.add_page(ExamplePage2(), last_page=True)


def test_example_wizard():
    """Test the import wizard"""
    # pylint: disable=import-outside-toplevel
    from guidata.qthelpers import qt_app_context

    with qt_app_context():
        wizard = ExampleWizard()
        wizard.exec()


if __name__ == "__main__":
    test_example_wizard()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Module checking dependencies with respect to a reference
"""

import hashlib
import os
import os.path as osp
import sys

from guidata.utils.misc import get_module_path

BUFFER_SIZE = 65536


def get_file_hash(filename, dhash=None, bufsize=None):
    """Return SHA256 hash for file"""
    if dhash is None:
        dhash = hashlib.sha256()
    buffer = bytearray(128 * 1024 if bufsize is None else bufsize)
    # using a memoryview so that we can slice the buffer without copying it
    buffer_view = memoryview(buffer)
    with open(filename, "rb", buffering=0) as fobj:
        while True:
            n_chunk = fobj.readinto(buffer_view)
            if not n_chunk:
                break
            dhash.update(buffer_view[:n_chunk])
    return dhash


def get_directory_hash(directory, extensions=None, verbose=False):
    """Return SHA256 hash for whole directory"""
    dhash = hashlib.sha1()
    if not os.path.exists(directory):
        return -1

    assert extensions is None or isinstance(extensions, (tuple, list))
    extlist = []
    if extensions is not None:
        for ext in extensions:
            if not ext.startswith("."):
                ext = "." + ext
            extlist.append(ext)
    filenb = 0
    for root, _dirs, files in os.walk(directory):
        for name in files:
            if extensions is None or osp.splitext(name)[1] in extlist:
                filepath = os.path.join(root, name)
                if verbose:
                    filenb += 1
                    print(f"[{filenb:07d}] Hashing: {filepath}")
                dhash = get_file_hash(filepath, dhash)
    return dhash.hexdigest()


def get_module_hash(module_name, verbose=False):
    """Return SHA256 hash for Python module"""
    path = get_module_path(module_name)
    return get_directory_hash(path, extensions=(".py",), verbose=verbose)


def get_dependencies_hash(dependencies):
    """Return SHA256 hash for Python module dependencies"""
    return {name: get_module_hash(name) for name in dependencies}


DEPFILENAME = f"dependencies-py{sys.version_info.major}-{sys.platform}.txt"


def check_dependencies_hash(datapath, hash_dict=None):
    """Check dependencies hash dictionary"""
    if hash_dict is None:
        hash_dict = {}
        with open(osp.join(datapath, DEPFILENAME), "rb") as filehandle:
            for line in filehandle.readlines():
                key, value = line.decode("utf-8").strip().split(":")
                hash_dict[key] = value
    return {name: get_module_hash(name) == hash_dict[name] for name in hash_dict}


def create_dependencies_file(datapath, dependencies):
    """Create dependencies.txt file"""
    hdict = get_dependencies_hash(dependencies)
    path = osp.join(datapath, DEPFILENAME)
    print("Hash file for Python module dependencies:")
    with open(path, "wb") as filehandle:
        print(f"  Opening {path}")
        for name in hdict:
            line = name + ":" + hdict[name] + os.linesep
            print(f"  Writing: {line}", end="")
            filehandle.write(line.encode("utf-8"))
        print("  => All done.")


if __name__ == "__main__":
    # datapath = osp.join(osp.dirname(__file__), os.pardir, "data")
    # hdict = get_dependencies_hash(("guidata", "plotpy"))
    # print(hdict)
    # print(check_dependencies_hash(datapath, hdict))
    create_dependencies_file(
        osp.join(osp.dirname(__file__), os.pardir, "data"), ("guidata", "plotpy")
    )
    # print(check_dependencies_hash(datapath))



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Test data functions

Functions creating test data: curves, images, ...
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: skip

from __future__ import annotations

from typing import Generator

import guidata.dataset as gds
import numpy as np

import cdl.obj
from cdl.algorithms.datatypes import is_integer_dtype
from cdl.config import _
from cdl.utils.tests import get_test_fnames


def get_test_signal(filename: str) -> cdl.obj.SignalObj:
    """Return test signal

    Args:
        filename: Filename

    Returns:
        Signal object
    """
    return cdl.obj.read_signal(get_test_fnames(filename)[0])


def get_test_image(filename: str) -> cdl.obj.ImageObj:
    """Return test image

    Args:
        filename: Filename

    Returns:
        Image object
    """
    return cdl.obj.read_image(get_test_fnames(filename)[0])


def create_paracetamol_signal(
    size: int | None = None, title: str | None = None
) -> cdl.obj.SignalObj:
    """Create test signal (Paracetamol molecule spectrum)

    Args:
        size: Size of the data. Defaults to None.
        title: Title of the signal. Defaults to None.

    Returns:
        Signal object
    """
    obj = cdl.obj.read_signal(get_test_fnames("paracetamol.txt")[0])
    if title is not None:
        obj.title = title
    if size is not None:
        x0, y0 = obj.xydata
        x1 = np.linspace(x0[0], x0[-1], size)
        y1 = np.interp(x1, x0, y0)
        obj.set_xydata(x1, y1)
    return obj


class GaussianNoiseParam(gds.DataSet):
    """Gaussian noise parameters"""

    mu = gds.FloatItem(
        _("Mean"),
        default=0.0,
        min=-100.0,
        max=100.0,
        help=_("Mean of the Gaussian distribution"),
    )
    sigma = gds.FloatItem(
        _("Standard deviation"),
        default=0.1,
        min=0.0,
        max=100.0,
        help=_("Standard deviation of the Gaussian distribution"),
    )
    seed = gds.IntItem(
        _("Seed"),
        default=1,
        min=0,
        max=1000000,
        help=_("Seed for random number generator"),
    )


def add_gaussian_noise_to_signal(
    signal: cdl.obj.SignalObj, p: GaussianNoiseParam | None = None
) -> None:
    """Add Gaussian (Normal-law) random noise to data

    Args:
        signal: Signal object
        p: Gaussian noise parameters.
    """
    if p is None:
        p = GaussianNoiseParam()
    rng = np.random.default_rng(p.seed)
    signal.data += rng.normal(p.mu, p.sigma, size=signal.data.shape)
    signal.title = f"GaussNoise({signal.title}, µ={p.mu}, σ={p.sigma})"


def create_noisy_signal(
    noiseparam: GaussianNoiseParam | None = None,
    newparam: cdl.obj.NewSignalParam | None = None,
    addparam: cdl.obj.GaussLorentzVoigtParam | None = None,
    title: str | None = None,
    noised: bool | None = None,
) -> cdl.obj.SignalObj:
    """Create curve data, optionally noised

    Args:
        noiseparam: Noise parameters. Default: None: No noise
        newparam: New signal parameters.
         Default: Gaussian, size=500, xmin=-10, xmax=10
        addparam: Additional parameters.
         Default: a=1.0, sigma=1.0, mu=0.0, ymin=0.0
        title: Title of the signal. Default: None
         If not None, overrides the title in newparam
        noised: If True, add noise to the signal.
         Default: None (use noiseparam)
         If True, eventually creates a new noiseparam if None

    Returns:
        cdl.obj.Signal object
    """
    if newparam is None:
        newparam = cdl.obj.NewSignalParam()
        newparam.stype = cdl.obj.SignalTypes.GAUSS
    if title is not None:
        newparam.title = title
    newparam.title = "Test signal (noisy)" if newparam.title is None else newparam.title
    if addparam is None:
        addparam = cdl.obj.GaussLorentzVoigtParam()
    if noised is not None and noised and noiseparam is None:
        noiseparam = GaussianNoiseParam()
        noiseparam.sigma = 5.0
    sig = cdl.obj.create_signal_from_param(newparam, addparam)
    if noiseparam is not None:
        add_gaussian_noise_to_signal(sig, noiseparam)
    return sig


def create_periodic_signal(
    shape: cdl.obj.SignalTypes,
    freq: float = 50.0,
    size: int = 10000,
    xmin: float = -10.0,
    xmax: float = 10.0,
) -> cdl.obj.SignalObj:
    """Create a periodic signal

    Args:
        shape: Shape of the signal
        freq: Frequency of the signal. Defaults to 50.0.
        size: Size of the signal. Defaults to 10000.
        xmin: Minimum value of the signal. Defaults to None.
        xmax: Maximum value of the signal. Defaults to None.

    Returns:
        Signal object
    """
    newparam = cdl.obj.new_signal_param(stype=shape, size=size, xmin=xmin, xmax=xmax)
    addparam = cdl.obj.PeriodicParam.create(freq=freq)
    return cdl.obj.create_signal_from_param(newparam, addparam)


def create_2d_steps_data(size: int, width: int, dtype: np.dtype) -> np.ndarray:
    """Creating 2D steps data for testing purpose

    Args:
        size: Size of the data
        width: Width of the steps
        dtype: Data type

    Returns:
        2D data
    """
    data = np.zeros((size, size), dtype=dtype)
    value = 1
    for col in range(0, size - width + 1, width):
        data[:, col : col + width] = np.array(value).astype(dtype)
        value *= 10
    data2 = np.zeros_like(data)
    value = 1
    for row in range(0, size - width + 1, width):
        data2[row : row + width, :] = np.array(value).astype(dtype)
        value *= 10
    data += data2
    return data


def create_2d_random(
    size: int, dtype: np.dtype, level: float = 0.1, seed: int = 1
) -> np.ndarray:
    """Creating 2D Uniform-law random image

    Args:
        size: Size of the data
        dtype: Data type
        level: Level of the random noise. Defaults to 0.1.
        seed: Seed for random number generator. Defaults to 1.

    Returns:
        2D data
    """
    rng = np.random.default_rng(seed)
    amp = (np.iinfo(dtype).max if is_integer_dtype(dtype) else 1.0) * level
    return np.array(rng.random((size, size)) * amp, dtype=dtype)


def create_2d_gaussian(
    size: int,
    dtype: np.dtype,
    x0: float = 0,
    y0: float = 0,
    mu: float = 0.0,
    sigma: float = 2.0,
    amp: float | None = None,
) -> np.ndarray:
    """Creating 2D Gaussian (-10 <= x <= 10 and -10 <= y <= 10)

    Args:
        size: Size of the data
        dtype: Data type
        x0: x0. Defaults to 0.
        y0: y0. Defaults to 0.
        mu: mu. Defaults to 0.0.
        sigma: sigma. Defaults to 2.0.
        amp: Amplitude. Defaults to None.

    Returns:
        2D data
    """
    xydata = np.linspace(-10, 10, size)
    x, y = np.meshgrid(xydata, xydata)
    if amp is None:
        try:
            amp = np.iinfo(dtype).max * 0.5
        except ValueError:
            # dtype is not integer
            amp = 1.0
    return np.array(
        amp
        * np.exp(
            -((np.sqrt((x - x0) ** 2 + (y - y0) ** 2) - mu) ** 2) / (2.0 * sigma**2)
        ),
        dtype=dtype,
    )


def get_laser_spot_data() -> list[np.ndarray]:
    """Return a list of NumPy arrays containing images which are relevant for
    testing laser spot image processing features

    Returns:
        List of NumPy arrays
    """
    znoise = create_2d_random(2000, np.uint16)
    zgauss = create_2d_gaussian(2000, np.uint16, x0=2.0, y0=-3.0)
    return [zgauss + znoise] + [
        cdl.obj.read_image(fname).data for fname in get_test_fnames("*.scor-data")
    ]


class PeakDataParam(gds.DataSet):
    """Peak data test image parameters"""

    size = gds.IntItem(_("Size"), default=2000, min=1)
    n_points = gds.IntItem(_("Number"), default=4, min=1, help=_("Number of points"))
    sigma_gauss2d = gds.FloatItem(
        "σ<sub>Gauss2D</sub>", default=0.06, help=_("Sigma of the 2D Gaussian")
    )
    amp_gauss2d = gds.IntItem(
        "A<sub>Gauss2D</sub>", default=1900, help=_("Amplitude of the 2D Gaussian")
    )
    mu_noise = gds.IntItem(
        "μ<sub>noise</sub>", default=845, help=_("Mean of the Gaussian distribution")
    )
    sigma_noise = gds.IntItem(
        "σ<sub>noise</sub>",
        default=25,
        help=_("Standard deviation of the Gaussian distribution"),
    )
    dx0 = gds.FloatItem("dx0", default=0.0)
    dy0 = gds.FloatItem("dy0", default=0.0)
    att = gds.FloatItem(_("Attenuation"), default=1.0)


def get_peak2d_data(
    p: PeakDataParam | None = None, seed: int | None = None, multi: bool = False
) -> np.ndarray:
    """Return a list of NumPy arrays containing images which are relevant for
    testing 2D peak detection or similar image processing features

    Args:
        p: Peak data test image parameters. Defaults to None.
        seed: Seed for random number generator. Defaults to None.
        multi: If True, multiple peaks are generated.
         Defaults to False.

    Returns:
        2D data
    """
    if p is None:
        p = PeakDataParam()
    delta = 0.1
    rng = np.random.default_rng(seed)
    coords = (rng.random((p.n_points, 2)) - 0.5) * 10 * (1 - delta)
    data = rng.normal(p.mu_noise, p.sigma_noise, size=(p.size, p.size))
    multi_nb = 2 if multi else 1
    for x0, y0 in coords:
        for idx in range(multi_nb):
            if idx != 0:
                p.dx0 = 0.08 + rng.random() * 0.08
                p.dy0 = 0.08 + rng.random() * 0.08
                p.att = 0.2 + rng.random() * 0.8
            data += create_2d_gaussian(
                p.size,
                np.uint16,
                x0=x0 + p.dx0,
                y0=y0 + p.dy0,
                sigma=p.sigma_gauss2d,
                amp=p.amp_gauss2d / multi_nb * p.att,
            )
    return data


def __set_default_size_dtype(
    p: cdl.obj.NewImageParam | None = None,
) -> cdl.obj.NewImageParam:
    """Set default shape and dtype

    Args:
        p: Image parameters. Defaults to None. If None, a new object is created.

    Returns:
        Image parameters
    """
    if p is None:
        p = cdl.obj.NewImageParam()
    p.height = 2000 if p.height is None else p.height
    p.width = 2000 if p.width is None else p.width
    p.dtype = cdl.obj.ImageDatatypes.UINT16 if p.dtype is None else p.dtype
    return p


def add_gaussian_noise_to_image(
    image: cdl.obj.ImageObj, param: cdl.obj.NormalRandomParam
) -> None:
    """Add Gaussian noise to image

    Args:
        src: Source image
        param: Parameters for the normal distribution
    """
    newparam = cdl.obj.new_image_param(
        height=image.data.shape[0],
        width=image.data.shape[1],
        dtype=cdl.obj.ImageDatatypes.from_dtype(image.data.dtype),
        itype=cdl.obj.ImageTypes.NORMALRANDOM,
    )
    noise = cdl.obj.create_image_from_param(newparam, param)
    image.data = image.data + noise.data


def create_checkerboard(
    p: cdl.obj.NewImageParam | None = None, num_checkers=8
) -> cdl.obj.ImageObj:
    """Generate a checkerboard pattern

    Args:
        p: Image parameters. Defaults to None.
        num_checkers: Number of checkers. Defaults to 8.
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (checkerboard)" if p.title is None else p.title
    obj = cdl.obj.create_image_from_param(p)
    re = np.r_[num_checkers * [0, 1]]  # one row of the checkerboard
    board = np.vstack(num_checkers * (re, re ^ 1))  # build the checkerboard
    board = np.kron(
        board, np.ones((p.height // num_checkers, p.height // num_checkers))
    )  # scale up the board
    obj.data = board
    return obj


def create_2dstep_image(
    p: cdl.obj.NewImageParam | None = None,
) -> cdl.obj.ImageObj:
    """Creating 2D step image

    Args:
        p: Image parameters. Defaults to None.

    Returns:
        Image object
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (2D step)" if p.title is None else p.title
    obj = cdl.obj.create_image_from_param(p)
    obj.data = create_2d_steps_data(p.height, p.height // 10, p.dtype.value)
    return obj


class RingParam(gds.DataSet):
    """Parameters for creating a ring image"""

    size = gds.IntItem(_("Size"), default=1000)
    ring_x0 = gds.IntItem(_("X<sub>center</sub>"), default=500)
    ring_y0 = gds.IntItem(_("Y<sub>center</sub>"), default=500)
    ring_width = gds.IntItem(_("Width"), default=10)
    ring_radius = gds.IntItem(_("Radius"), default=250)
    ring_intensity = gds.IntItem(_("Intensity"), default=1000)


def create_ring_data(
    size: int, x0: int, y0: int, width: int, radius: int, intensity: int
) -> np.ndarray:
    """Create 2D ring data

    Args:
        size: Size of the image
        x0: Center x coordinate
        y0: Center y coordinate
        width: Width of the ring
        radius: Radius of the ring
        intensity: Intensity of the ring

    Returns:
        2D data
    """
    data = np.zeros((size, size), dtype=np.uint16)
    for x in range(data.shape[0]):
        for y in range(data.shape[1]):
            if (x - x0) ** 2 + (y - y0) ** 2 >= (radius - width) ** 2 and (
                x - x0
            ) ** 2 + (y - y0) ** 2 <= (radius + width) ** 2:
                data[x, y] = intensity
    return data


def create_ring_image(p: RingParam | None = None) -> cdl.obj.ImageObj:
    """Creating 2D ring image

    Args:
        p: Ring image parameters. Defaults to None.

    Returns:
        Image object
    """
    if p is None:
        p = RingParam()
    obj = cdl.obj.create_image(
        f"Ring(size={p.size},x0={p.ring_x0},y0={p.ring_y0},width={p.ring_width},"
        f"radius={p.ring_radius},intensity={p.ring_intensity})"
    )
    obj.data = create_ring_data(
        p.size,
        p.ring_x0,
        p.ring_y0,
        p.ring_width,
        p.ring_radius,
        p.ring_intensity,
    )
    return obj


def create_peak2d_image(
    p: cdl.obj.NewImageParam | None = None,
) -> cdl.obj.ImageObj:
    """Creating 2D peak image

    Args:
        p: Image parameters. Defaults to None.

    Returns:
        Image object
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (2D peaks)" if p.title is None else p.title
    obj = cdl.obj.create_image_from_param(p)
    param = PeakDataParam()
    if p.height is not None and p.width is not None:
        param.size = max(p.height, p.width)
    obj.data = get_peak2d_data(param)
    return obj


def create_sincos_image(
    p: cdl.obj.NewImageParam | None = None,
) -> cdl.obj.ImageObj:
    """Creating test image (sin(x)+cos(y))

    Args:
        p: Image parameters. Defaults to None.

    Returns:
        Image object
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (sin(x)+cos(y))" if p.title is None else p.title
    dtype = p.dtype.value
    x, y = np.meshgrid(np.linspace(0, 10, p.width), np.linspace(0, 10, p.height))
    raw_data = 0.5 * (np.sin(x) + np.cos(y)) + 0.5
    obj = cdl.obj.create_image_from_param(p)
    if np.issubdtype(dtype, np.floating):
        obj.data = raw_data
        return obj
    dmin = np.iinfo(dtype).min * 0.95
    dmax = np.iinfo(dtype).max * 0.95
    obj.data = np.array(raw_data * (dmax - dmin) + dmin, dtype=dtype)
    return obj


def create_noisygauss_image(
    p: cdl.obj.NewImageParam | None = None,
    center: tuple[float, float] | None = None,
    level: float = 0.1,
    add_annotations: bool = False,
) -> cdl.obj.ImageObj:
    """Create test image (2D noisy gaussian)

    Args:
        p: Image parameters. Defaults to None.
        center: Center of the gaussian. Defaults to None.
        level: Level of the random noise. Defaults to 0.1.
        add_annotations: If True, add annotations. Defaults to False.

    Returns:
        Image object
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (noisy 2D Gaussian)" if p.title is None else p.title
    dtype = p.dtype.value
    size = max(p.height, p.width)
    obj = cdl.obj.create_image_from_param(p)
    if center is None:
        # Default center
        x0, y0 = 2.0, 3.0
    else:
        x0, y0 = center
    obj.data = create_2d_gaussian(size, dtype=dtype, x0=x0, y0=y0)
    if level:
        obj.data += create_2d_random(size, dtype, level)
    if add_annotations:
        obj.add_annotations_from_file(get_test_fnames("annotations.json")[0])
    if p.height != p.width:
        if p.height > p.width:
            obj.data = obj.data[:, p.width // 2 : -p.width // 2]
        else:
            obj.data = obj.data[p.height // 2 : -p.height // 2, :]
    return obj


def create_multigauss_image(
    p: cdl.obj.NewImageParam | None = None,
) -> cdl.obj.ImageObj:
    """Create test image (multiple 2D-gaussian peaks)

    Args:
        p: Image parameters. Defaults to None.

    Returns:
        Image object
    """
    p = __set_default_size_dtype(p)
    p.title = "Test image (multi-2D-gaussian)" if p.title is None else p.title
    dtype = p.dtype.value
    size = p.width
    obj = cdl.obj.create_image_from_param(p)
    obj.data = (
        create_2d_gaussian(size, dtype, x0=0.5, y0=3.0)
        + create_2d_gaussian(size, dtype, x0=-1.0, y0=-1.0, sigma=1.0)
        + create_2d_gaussian(size, dtype, x0=7.0, y0=8.0)
    )
    return obj


def create_annotated_image(title: str | None = None) -> cdl.obj.ImageObj:
    """Create test image with annotations

    Returns:
        Image object
    """
    data = create_2d_gaussian(600, np.uint16, x0=2.0, y0=3.0)
    title = "Test image (with metadata)" if title is None else title
    image = cdl.obj.create_image(title, data)
    image.add_annotations_from_file(get_test_fnames("annotations.json")[0])
    return image


def create_resultshapes() -> Generator[cdl.obj.ResultShape, None, None]:
    """Create test result shapes (core.model.base.ResultShape test objects)

    Yields:
        ResultShape object
    """
    for shape, data in (
        ("circle", [[0, 250, 250, 200], [0, 250, 250, 140]]),
        ("rectangle", [0, 300, 200, 700, 700]),
        ("segment", [0, 50, 250, 400, 400]),
        ("point", [[0, 500, 500], [0, 15, 400]]),
        (
            "polygon",
            [0, 100, 100, 150, 100, 150, 150, 200, 100, 250, 50],
        ),
    ):
        yield cdl.obj.ResultShape(shape, data, shape, add_label=shape == "segment")


def create_resultproperties() -> Generator[cdl.obj.ResultProperties, None, None]:
    """Create test result properties (core.model.base.ResultProperties test object)

    Returns:
        ResultProperties object
    """
    for title, data, labels in (
        ("TestProperties1", [0, 2.5, -30909, 1.0, 0.0], ["A", "B", "C", "D"]),
        (
            "TestProperties2",
            [[0, 1.232325, -9, 0, 10], [0, 250, -3, 12.0, 530.0]],
            ["P1", "P2", "P3", "P4"],
        ),
    ):
        yield cdl.obj.ResultProperties(title, data, labels)



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.


"""
DataLab HDF5 importer module
"""

# pylint: disable=unused-import

# Registering dynamic I/O features:
from cdl.core.io.h5 import generic  # noqa: F401
from cdl.core.io.h5.common import H5Importer  # noqa: F401



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Application test for main window
--------------------------------

Testing the features of the main window of the application that are not
covered by other tests.
"""

# guitest: show

import os

from cdl.env import execenv
from cdl.param import MovingMedianParam
from cdl.tests import cdltest_app_context
from cdl.tests.data import create_paracetamol_signal


def test_main_app():
    """Main window test"""
    with cdltest_app_context(console=False) as win:
        # Switch from panel to panel
        for panelname in ("macro", "image", "signal"):
            win.set_current_panel(panelname)
        # Switch to an unknown panel
        try:
            win.set_current_panel("unknown_panel")
            raise RuntimeError("Unknown panel should have raised an exception")
        except ValueError:
            pass

        panel = win.signalpanel

        # Create new groups
        panel.add_group("Group 1")
        panel.add_group("Group 2")
        # Rename group
        panel.objview.select_groups([2])
        panel.rename_group("Group xxx")
        panel.remove_object(force=True)

        # Add signals to signal panel
        sig1 = create_paracetamol_signal(500)
        panel.add_object(sig1)
        panel.processor.compute_derivative()
        panel.processor.compute_wiener()

        # Get object titles
        titles = win.get_object_titles()
        execenv.print(f"Object titles:{os.linesep}{titles}")

        # Get object uuids
        uuids = win.get_object_uuids()
        execenv.print(f"Object uuids:{os.linesep}{uuids}")

        # Testing `get_object`
        execenv.print("*** Testing `get_object` ***")
        # Get object from title
        obj = win.get_object(titles[-1])
        execenv.print(f"  Object (from title) '{obj.short_id}':{os.linesep}{obj}")
        # Get object
        obj = win.get_object(1)
        execenv.print(f"  Object (from number)  '{obj.short_id}':{os.linesep}{obj}")
        # Get object by uuid
        obj = win.get_object(uuids[-1])
        execenv.print(f"  Object (from uuid)  '{obj.short_id}':{os.linesep}{obj}")

        # Testing dict-like interface of main window:
        execenv.print("*** Testing dict-like interface of proxy ***")
        # Get object from title
        obj = win[titles[-1]]
        execenv.print(f"  Object (from title) '{obj.short_id}':{os.linesep}{obj}")
        # Get object
        obj = win[1]
        execenv.print(f"  Object (from number)  '{obj.short_id}':{os.linesep}{obj}")
        # Get object by uuid
        obj = win[uuids[-1]]
        execenv.print(f"  Object (from uuid)  '{obj.short_id}':{os.linesep}{obj}")

        # Use "calc" method with parameters
        param = MovingMedianParam.create(n=5)
        win.calc("compute_moving_median", param)
        # Use "calc" method without parameters
        win.calc("compute_integral")
        # Use "calc" and choose an unknown computation method
        try:
            win.calc("unknown_method")
            raise RuntimeError("Unknown method should have raised an exception")
        except ValueError:
            pass

        # Force application menus to pop-up
        for menu in (
            win.file_menu,
            win.edit_menu,
            win.operation_menu,
            win.processing_menu,
            win.analysis_menu,
            win.view_menu,
            win.help_menu,
        ):
            menu.popup(menu.pos())
        win.file_menu.popup(win.mapToGlobal(win.file_menu.pos()))

        # Open settings dialog
        win.settings_action.trigger()


if __name__ == "__main__":
    test_main_app()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image FFT unit test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# pylint: disable=duplicate-code
# guitest: show

import numpy as np
import pytest
from guidata.qthelpers import qt_app_context

import cdl.algorithms.image as alg
import cdl.computation.image as cpi
import cdl.tests.data as ctd
from cdl.env import execenv
from cdl.utils.tests import check_array_result, check_scalar_result
from cdl.utils.vistools import view_images_side_by_side


def test_image_fft_interactive():
    """2D FFT interactive test."""
    with qt_app_context():
        # Create a 2D ring image
        execenv.print("Generating 2D ring image...", end=" ")
        data = ctd.create_ring_image(ctd.RingParam()).data
        execenv.print("OK")

        # FFT
        execenv.print("Computing FFT of image...", end=" ")
        f = alg.fft2d(data)
        data2 = alg.ifft2d(f)
        execenv.print("OK")
        execenv.print("Comparing original and FFT/iFFT images...", end=" ")
        check_array_result(
            "Image FFT/iFFT", np.array(data2.real, dtype=data.dtype), data, rtol=1e-3
        )
        execenv.print("OK")

        images = [data, f.real, f.imag, np.abs(f), data2.real, data2.imag]
        titles = ["Original", "Re(FFT)", "Im(FFT)", "Abs(FFT)", "Re(iFFT)", "Im(iFFT)"]
        view_images_side_by_side(images, titles, rows=2, title="2D FFT/iFFT")


@pytest.mark.validation
def test_image_fft() -> None:
    """2D FFT validation test."""
    ima1 = ctd.create_checkerboard()
    fft = cpi.compute_fft(ima1)
    ifft = cpi.compute_ifft(fft)

    # Check that the inverse FFT reconstructs the original image
    check_array_result("Checkerboard image FFT/iFFT", ifft.data.real, ima1.data)

    # Parseval's Theorem Validation
    original_energy = np.sum(np.abs(ima1.data) ** 2)
    transformed_energy = np.sum(np.abs(fft.data) ** 2) / (ima1.data.size)
    check_scalar_result("Parseval's Theorem", transformed_energy, original_energy)


@pytest.mark.skip(reason="Already covered by the `test_image_fft` test.")
@pytest.mark.validation
def test_image_ifft() -> None:
    """2D iFFT validation test."""
    # This is just a way of marking the iFFT test as a validation test because it is
    # already covered by the FFT test above (there is no need to repeat the same test).


@pytest.mark.validation
def test_image_magnitude_spectrum() -> None:
    """2D magnitude spectrum validation test."""
    ima1 = ctd.create_checkerboard()
    fft = cpi.compute_fft(ima1)
    mag = cpi.compute_magnitude_spectrum(ima1)

    # Check that the magnitude spectrum is correct
    exp = np.abs(fft.data)
    check_array_result("Checkerboard image FFT magnitude spectrum", mag.data, exp)


@pytest.mark.validation
def test_image_phase_spectrum() -> None:
    """2D phase spectrum validation test."""
    ima1 = ctd.create_checkerboard()
    fft = cpi.compute_fft(ima1)
    phase = cpi.compute_phase_spectrum(ima1)

    # Check that the phase spectrum is correct
    exp = np.rad2deg(np.angle(fft.data))
    check_array_result("Checkerboard image FFT phase spectrum", phase.data, exp)


@pytest.mark.validation
def test_image_psd() -> None:
    """2D Power Spectral Density validation test."""
    ima1 = ctd.create_checkerboard()
    psd = cpi.compute_psd(ima1)

    # Check that the PSD is correct
    exp = np.abs(cpi.compute_fft(ima1).data) ** 2
    check_array_result("Checkerboard image PSD", psd.data, exp)


if __name__ == "__main__":
    test_image_fft_interactive()
    test_image_fft()
    test_image_magnitude_spectrum()
    test_image_phase_spectrum()
    test_image_psd()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Contour finding application test
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

import cdl.obj
import cdl.param
from cdl.env import execenv
from cdl.tests import cdltest_app_context, take_plotwidget_screenshot
from cdl.tests.data import create_multigauss_image
from cdl.tests.features.common.roi_app_test import create_test_image_with_roi


def test_contour_app():
    """Run contour finding application test scenario"""
    newparam = cdl.obj.new_image_param(height=200, width=200)
    with cdltest_app_context() as win:
        panel = win.imagepanel
        for shape in ("polygon", "circle", "ellipse"):
            ima1 = create_multigauss_image(newparam)
            ima1.metadata["colormap"] = "gray"
            panel.add_object(ima1)
            param = cdl.param.ContourShapeParam.create(shape=shape)
            panel.processor.compute_contour_shape(param)
            take_plotwidget_screenshot(panel, "contour_test")
            ima2 = create_test_image_with_roi(newparam)
            ima2.metadata["colormap"] = "gray"
            panel.add_object(ima2)
            panel.processor.edit_regions_of_interest()
            panel.processor.compute_contour_shape(param)
            if not execenv.unattended:
                break


if __name__ == "__main__":
    test_contour_app()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
.. Signal panel (see parent package :mod:`cdl.core.gui.panel`)
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...

from __future__ import annotations

from typing import TYPE_CHECKING, Type

from plotpy.tools import (
    HCursorTool,
    HRangeTool,
    LabelTool,
    RectangleTool,
    SegmentTool,
    VCursorTool,
    XCursorTool,
)

from cdl.config import _
from cdl.core.gui import roieditor
from cdl.core.gui.actionhandler import SignalActionHandler
from cdl.core.gui.panel.base import BaseDataPanel
from cdl.core.gui.plothandler import SignalPlotHandler
from cdl.core.gui.processor.signal import SignalProcessor
from cdl.core.io.signal import SignalIORegistry
from cdl.core.model.signal import (
    CURVESTYLES,
    SignalObj,
    SignalROI,
    create_signal_from_param,
    new_signal_param,
)

if TYPE_CHECKING:
    import guidata.dataset as gds
    from qtpy import QtWidgets as QW

    from cdl.core.gui.docks import DockablePlotWidget
    from cdl.core.model.signal import NewSignalParam


class SignalPanel(BaseDataPanel[SignalObj, SignalROI, roieditor.SignalROIEditor]):
    """Object handling the item list, the selected item properties and plot,
    specialized for Signal objects"""

    PANEL_STR = _("Signal Panel")
    PANEL_STR_ID = "signal"
    PARAMCLASS = SignalObj

    # The following tools are used to create annotations on signals. The annotation
    # items are created using PlotPy's default settings. Those appearance settings
    # may be modified in the configuration (see `cdl.config`).
    ANNOTATION_TOOLS = (
        LabelTool,
        VCursorTool,
        HCursorTool,
        XCursorTool,
        SegmentTool,
        RectangleTool,
        HRangeTool,
    )

    IO_REGISTRY = SignalIORegistry
    H5_PREFIX = "DataLab_Sig"

    # pylint: disable=duplicate-code

    @staticmethod
    def get_roieditor_class() -> Type[roieditor.SignalROIEditor]:
        """Return ROI editor class"""
        return roieditor.SignalROIEditor

    def __init__(
        self,
        parent: QW.QWidget,
        dockableplotwidget: DockablePlotWidget,
        panel_toolbar: QW.QToolBar,
    ) -> None:
        super().__init__(parent)
        self.plothandler = SignalPlotHandler(self, dockableplotwidget.plotwidget)
        self.processor = SignalProcessor(self, dockableplotwidget.plotwidget)
        view_toolbar = dockableplotwidget.toolbar
        self.acthandler = SignalActionHandler(self, panel_toolbar, view_toolbar)

    # ------Creating, adding, removing objects------------------------------------------
    def get_newparam_from_current(
        self, newparam: NewSignalParam | None = None, title: str | None = None
    ) -> NewSignalParam | None:
        """Get new object parameters from the current object.

        Args:
            newparam (guidata.dataset.DataSet): new object parameters.
             If None, create a new one.
            title: new object title. If None, use the current object title, or the
             default title.

        Returns:
            New object parameters
        """
        curobj: SignalObj = self.objview.get_current_object()
        if newparam is None:
            newparam = new_signal_param(title=title)
        if curobj is not None:
            newparam.size = len(curobj.data)
            newparam.xmin = curobj.x.min()
            newparam.xmax = curobj.x.max()
        return newparam

    def new_object(
        self,
        newparam: NewSignalParam | None = None,
        addparam: gds.DataSet | None = None,
        edit: bool = True,
        add_to_panel: bool = True,
    ) -> SignalObj | None:
        """Create a new object (signal).

        Args:
            newparam (guidata.dataset.DataSet): new object parameters
            addparam (guidata.dataset.DataSet): additional parameters
            edit (bool): Open a dialog box to edit parameters (default: True)
            add_to_panel (bool): Add the new object to the panel (default: True)

        Returns:
            New object
        """
        if not self.mainwindow.confirm_memory_state():
            return None
        newparam = self.get_newparam_from_current(newparam)
        signal = create_signal_from_param(
            newparam, addparam=addparam, edit=edit, parent=self.parent()
        )
        if signal is None:
            return None
        if add_to_panel:
            self.add_object(signal)
        return signal

    # ------Plotting--------------------------------------------------------------------
    def toggle_anti_aliasing(self, state: bool) -> None:
        """Toggle anti-aliasing on/off

        Args:
            state: state of the anti-aliasing
        """
        self.plothandler.toggle_anti_aliasing(state)

    def reset_curve_styles(self) -> None:
        """Reset curve styles"""
        CURVESTYLES.reset_styles()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Parameters (:mod:`cdl.param`)
-----------------------------

The :mod:`cdl.param` module aims at providing all the dataset parameters that are used
by the :mod:`cdl.computation` and :mod:`cdl.core.gui.processor` packages.

Those datasets are defined in other modules:

    - :mod:`cdl.computation.base`
    - :mod:`cdl.computation.image`
    - :mod:`cdl.computation.signal`

The :mod:`cdl.param` module is thus a convenient way to import all the sets of
parameters at once.

As a matter of fact, the following import statement is equivalent to the previous one:

.. code-block:: python

    # Original import statement
    from cdl.computation.base import MovingAverageParam
    from cdl.computation.signal import PolynomialFitParam
    from cdl.computation.image.exposure import EqualizeHistParam

    # Equivalent import statement
    from cdl.param import MovingAverageParam, PolynomialFitParam, EqualizeHistParam

Introduction to `DataSet` parameters
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

The datasets listed in the following sections are used to define the parameters
necessary for the various computations and processing operations available in DataLab.

Each dataset is a subclass of :py:class:`guidata.dataset.datatypes.DataSet` and thus
needs to be instantiated before being used.

Here is a complete example of how to instantiate a dataset and access its parameters
with the :py:class:`cdl.param.BinningParam` dataset:

    .. autodataset:: cdl.param.BinningParam
        :no-index:
        :shownote:

Common parameters
^^^^^^^^^^^^^^^^^

.. autodataset:: cdl.param.ArithmeticParam
    :no-index:
.. autodataset:: cdl.param.ClipParam
    :no-index:
.. autodataset:: cdl.param.FFTParam
    :no-index:
.. autodataset:: cdl.param.SpectrumParam
    :no-index:
.. autodataset:: cdl.param.GaussianParam
    :no-index:
.. autodataset:: cdl.param.MovingAverageParam
    :no-index:
.. autodataset:: cdl.param.MovingMedianParam
    :no-index:
.. autodataset:: cdl.param.ConstantParam
    :no-index:

Signal parameters
^^^^^^^^^^^^^^^^^

.. autodataset:: cdl.param.DataTypeSParam
    :no-index:
.. autodataset:: cdl.param.FWHMParam
    :no-index:
.. autodataset:: cdl.param.NormalizeParam
    :no-index:
.. autodataset:: cdl.param.PeakDetectionParam
    :no-index:
.. autodataset:: cdl.param.PolynomialFitParam
    :no-index:
.. autodataset:: cdl.param.XYCalibrateParam
    :no-index:
.. autodataset:: cdl.param.InterpolationParam
    :no-index:
.. autodataset:: cdl.param.ResamplingParam
    :no-index:
.. autodataset:: cdl.param.DetrendingParam
    :no-index:
.. autodataset:: cdl.param.PowerParam
    :no-index:
.. autodataset:: cdl.param.WindowingParam
    :no-index:
.. autodataset:: cdl.param.LowPassFilterParam
    :no-index:
.. autodataset:: cdl.param.HighPassFilterParam
    :no-index:
.. autodataset:: cdl.param.BandPassFilterParam
    :no-index:
.. autodataset:: cdl.param.BandStopFilterParam
    :no-index:
.. autodataset:: cdl.param.DynamicParam
    :no-index:

Image parameters
^^^^^^^^^^^^^^^^

Base image parameters
~~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.BinningParam
    :no-index:
.. autodataset:: cdl.param.ButterworthParam
    :no-index:
.. autodataset:: cdl.param.DataTypeIParam
    :no-index:
.. autodataset:: cdl.param.FlatFieldParam
    :no-index:
.. autodataset:: cdl.param.GridParam
    :no-index:
.. autodataset:: cdl.param.HoughCircleParam
    :no-index:
.. autodataset:: cdl.param.LogP1Param
    :no-index:
.. autodataset:: cdl.param.LineProfileParam
    :no-index:
.. autodataset:: cdl.param.SegmentProfileParam
    :no-index:
.. autodataset:: cdl.param.AverageProfileParam
    :no-index:
.. autodataset:: cdl.param.RadialProfileParam
    :no-index:
.. autodataset:: cdl.param.ResizeParam
    :no-index:
.. autodataset:: cdl.param.RotateParam
    :no-index:
.. autodataset:: cdl.param.ZCalibrateParam
    :no-index:

Threshold parameters
~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.ThresholdParam
    :no-index:

Exposure correction parameters
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.AdjustGammaParam
    :no-index:
.. autodataset:: cdl.param.AdjustLogParam
    :no-index:
.. autodataset:: cdl.param.AdjustSigmoidParam
    :no-index:
.. autodataset:: cdl.param.EqualizeAdaptHistParam
    :no-index:
.. autodataset:: cdl.param.EqualizeHistParam
    :no-index:
.. autodataset:: cdl.param.RescaleIntensityParam
    :no-index:

Restoration parameters
~~~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.DenoiseBilateralParam
    :no-index:
.. autodataset:: cdl.param.DenoiseTVParam
    :no-index:
.. autodataset:: cdl.param.DenoiseWaveletParam
    :no-index:

Morphological parameters
~~~~~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.MorphologyParam
    :no-index:

Edge detection parameters
~~~~~~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.CannyParam
    :no-index:

Detection parameters
~~~~~~~~~~~~~~~~~~~~

.. autodataset:: cdl.param.BlobDOGParam
    :no-index:
.. autodataset:: cdl.param.BlobDOHParam
    :no-index:
.. autodataset:: cdl.param.BlobLOGParam
    :no-index:
.. autodataset:: cdl.param.BlobOpenCVParam
    :no-index:
.. autodataset:: cdl.param.ContourShapeParam
    :no-index:
.. autodataset:: cdl.param.Peak2DDetectionParam
    :no-index:
"""

# pylint:disable=unused-import
# flake8: noqa

from cdl.computation.base import (
    ArithmeticParam,
    ClipParam,
    FFTParam,
    SpectrumParam,
    GaussianParam,
    HistogramParam,
    MovingAverageParam,
    MovingMedianParam,
    ConstantParam,
    NormalizeParam,
)
from cdl.computation.image import (
    AverageProfileParam,
    BinningParam,
    ButterworthParam,
    DataTypeIParam,
    FlatFieldParam,
    GridParam,
    HoughCircleParam,
    LogP1Param,
    LineProfileParam,
    SegmentProfileParam,
    RadialProfileParam,
    ResizeParam,
    RotateParam,
    ZCalibrateParam,
)
from cdl.computation.image.detection import (
    BlobDOGParam,
    BlobDOHParam,
    BlobLOGParam,
    BlobOpenCVParam,
    ContourShapeParam,
    Peak2DDetectionParam,
)
from cdl.computation.image.edges import CannyParam
from cdl.computation.image.threshold import ThresholdParam
from cdl.computation.image.exposure import (
    AdjustGammaParam,
    AdjustLogParam,
    AdjustSigmoidParam,
    EqualizeAdaptHistParam,
    EqualizeHistParam,
    RescaleIntensityParam,
)
from cdl.computation.image.morphology import MorphologyParam
from cdl.computation.image.restoration import (
    DenoiseBilateralParam,
    DenoiseTVParam,
    DenoiseWaveletParam,
)
from cdl.computation.signal import (
    DataTypeSParam,
    DetrendingParam,
    FWHMParam,
    InterpolationParam,
    PeakDetectionParam,
    PolynomialFitParam,
    ResamplingParam,
    XYCalibrateParam,
    PowerParam,
    WindowingParam,
    LowPassFilterParam,
    HighPassFilterParam,
    BandPassFilterParam,
    BandStopFilterParam,
    DynamicParam,
)
